import warnings
import difflib
from typing import Union, Type
from collections.abc import Callable
import json

from .tools import tool_suggest_table, tool_dispatch_table, solving_methods


class InversionOptions:
    r"""Class for specification on how an inversion will run, including which backend
    tool to use and solver-specific parameters.

    .. tip::

        A typical workflow of :code:`InversionOptions`:

        Step 1 (optional): let the :ref:`Guidance Methods <guide>` to walk you through
        available solving methods in a hierarchical way.

        Step 2: Use the :ref:`Set/Unset Backend Tools <set_unset_tools>` to fix your choice
        on which backend tool to use.

        Step 3: Set solver-specific parameters using the :ref:`Solver Params <set_params>`
        related methods.


    .. admonition:: Example usage of InversionOptions
        :class: dropdown, tip

        >>> from cofi import InversionOptions
        >>> inv_options = InversionOptions()
        >>> inv_options.get_default_tool()
        'scipy.optimize.minimize'
        >>> inv_options.suggest_tools()
        Here's a complete list of inversion tools supported by CoFI (grouped by methods):
        {
            "optimization": [
                "scipy.optimize.minimize",
                "scipy.optimize.least_squares"
            ],
            "matrix solvers": [
                "scipy.linalg.lstsq"
            ],
            "sampling": [
                "emcee"
            ]
        }
        >>> inv_options.set_tool("scipy.linalg.lstsq")
        >>> inv_options.suggest_solver_params()
        Current backend tool scipy.linalg.lstsq has the following solver-specific parameters:
        Required parameters:
        -- nothing --
        Optional parameters & default settings:
        {'cond': None, 'overwrite_a': False, 'overwrite_b': False, 'check_finite': True, 'lapack_driver': None}
        >>> inv_options.summary()
        =============================
        Summary for inversion options
        =============================
        Solving method: None set
        Use `suggest_solving_methods()` to check available solving methods.
        -----------------------------
        Backend tool: `scipy.linalg.lstsq` - SciPy's wrapper function over LAPACK's
        linear least-squares solver, using 'gelsd', 'gelsy' (default), or 'gelss' as
        backend driver
        References: ['https://docs.scipy.org/doc/scipy/reference/generated/scipy.linalg.lstsq.html',
        'https://www.netlib.org/lapack/lug/node27.html']
        Use `suggest_tools()` to check available backend tools.
        -----------------------------
        Solver-specific parameters: None set
        Use `suggest_solver_params()` to check required/optional solver-specific parameters.

    .. _guide:

    .. rubric:: Guidance Methods

    Here are how you can walk through our solvers tree as a guidance. Note that this
    step is optional, and you can always jump to :ref:`Set/Unset Backend Tools <set_unset_tools>`
    directly.

    .. autosummary::
        InversionOptions.set_solving_method
        InversionOptions.unset_solving_method

    :ref:`back to top <top_InversionOptions>`

    .. _set_unset_tools:

    .. rubric:: Set/Unset Backend Tools

    To select/unselect a backend tool, use the following methods

    .. autosummary::
        InversionOptions.set_tool
        InversionOptions.unset_tool
        InversionOptions.get_tool
        InversionOptions.get_default_tool
        InversionOptions.suggest_tools

    :ref:`back to top <top_InversionOptions>`

    .. _set_params:

    .. rubric:: Solver Params

    To set tool-specific parameters, use the following methods

    .. autosummary::
        InversionOptions.set_params
        InversionOptions.get_params
        InversionOptions.suggest_solver_params

    :ref:`back to top <top_InversionOptions>`

    """

    def __init__(self):
        self.hyper_params = {}
        self.tool = None
        self.method = None

    def set_params(self, **kwargs):
        r"""Sets solver-specific parameters

        Use :func:`InversionOptions.suggest_solver_params` to get a list of parameters
        required and optional.

        To set the parameters, use argument keyword to specify directly which parameter
        you refer to.

        Examples
        --------

        .. admonition:: code example
            :class: dropdown, attention

            .. code-block:: pycon
                :emphasize-lines: 9

                >>> from cofi import InversionOptions
                >>> inv_options = InversionOptions()
                >>> inv_options.suggest_solver_params()
                Current backend tool scipy.optimize.minimize (default) has the following solver-specific parameters:
                Required parameters:
                -- nothing --
                Optional parameters & default settings:
                {'method': None, 'tol': None, 'callback': None, 'options': None}
                >>> inv_options.set_params(method="Nelder-Mead")

        """
        self.hyper_params.update(kwargs)

    def get_params(self) -> dict:
        r"""Get solver-specific parameters defined so far

        Returns
        -------
        dict
            a Python dictionary that maps solver-specific parameter name to the value
            you've set.

        Examples
        --------

        .. admonition:: code example
            :class: dropdown, attention

            .. code-block:: pycon
                :emphasize-lines: 4

                >>> from cofi import InversionOptions
                >>> inv_options = InversionOptions()
                >>> inv_options.set_params(method="Nelder-Mead")
                >>> inv_options.get_params()
                {'method': 'Nelder-Mead'}

        """
        return self.hyper_params

    def set_solving_method(self, method: str):
        r"""Sets the solving method

        .. warning::
            The current version is a flattened version of our inference tools tree, we
            are going to change this interface very soon.

        Use :func:`InversionOptions.suggest_solving_methods` to get a list of solving
        methods to choose from.

        Parameters
        ----------
        method : str
            the string that represents a solving approach

        Raises
        ------
        ValueError
            when the solving method you attempt to set is invalid

        Examples
        --------

        .. admonition:: code example
            :class: dropdown, attention

            .. code-block:: pycon
                :emphasize-lines: 3

                >>> from cofi import InversionOptions
                >>> inv_options = InversionOptions()
                >>> inv_options.set_solving_method("matrix solvers")
                >>> inv_options.suggest_tools()
                Based on the solving method you've set, the following tools are suggested:
                ['scipy.linalg.lstsq']

                Use `InversionOptions.set_tool(tool_name)` to set a specific tool from above
                Use `InversionOptions.set_solving_method(method_name)` to change solving method
                Use `InversionOptions.unset_solving_method()` if you'd like to see more options
                Check CoFI documentation 'Advanced Usage' section for how to plug in your own solver

        """
        if method is None:
            self.unset_solving_method()
        elif method in solving_methods:
            self.method = method
        else:
            close_matches = difflib.get_close_matches(method, solving_methods)
            _error_msg_suffix = (
                f"\n\nDid you mean '{close_matches[0]}'?" if len(close_matches) else ""
            )
            raise ValueError(
                "the solving method is invalid, please choose from"
                f" {solving_methods}{_error_msg_suffix}"
            )

    def unset_solving_method(self):
        """Unsets the chosen solving approach"""
        self.method = None

    def set_tool(self, tool: Union[str, Type]):
        r"""Sets the tool that will be the backend solver for your inversion problem

        This can be:

        - a backend tool we support, use :func:`InversionOptions.suggest_tools` to get
          a list of tools you can choose from
        - or your own solver class, check `our tutorial - Advanced Usage <tutorial.html#advanced-usage>`_
          for details about how to define and use your custom inference tool

        Parameters
        ----------
        tool : Union[str, Type]
            either the name of a backend tool or your custom :class:`tools.BaseInferenceTool`
            class

        Raises
        ------
        ValueError
            when the string you pass in isn't in our supported tools list, or when the
            inference tool class you pass in doesn't implement the ``__call__(self,)``
            method.

        Examples
        --------

        .. admonition:: code example: set a supported tool
            :class: dropdown, attention

            .. code-block:: pycon
                :emphasize-lines: 3

                >>> from cofi import InversionOptions
                >>> inv_options = InversionOptions()
                >>> inv_options.set_tool("scipy.linalg.lstsq")
                >>> inv_options.summary()
                =============================
                Summary for inversion options
                =============================
                Solving method: None set
                Use `suggest_solving_methods()` to check available solving methods.
                -----------------------------
                Backend tool: `scipy.linalg.lstsq` - SciPy's wrapper function over LAPACK's linear least-squares solver, using 'gelsd', 'gelsy' (default), or 'gelss' as backend driver
                References: ['https://docs.scipy.org/doc/scipy/reference/generated/scipy.linalg.lstsq.html', 'https://www.netlib.org/lapack/lug/node27.html']
                Use `suggest_tools()` to check available backend tools.
                -----------------------------
                Solver-specific parameters: None set
                Use `suggest_solver_params()` to check required/optional solver-specific parameters.


        .. admonition:: code example: set your own solver
            :class: dropdown, attention

            .. code-block:: pycon
                :emphasize-lines: 3-8, 11

                >>> from cofi.tools import BaseInferenceTool
                >>> from cofi import InversionOptions
                >>> class MyDummySolver(BaseInferenceTool):
                ...   short_description = "My dummy solver that always return (1,2) as result"
                ...   def __init__(self, inv_problem, inv_options):
                ...     super().__init__(inv_problem, inv_options)
                ...   def __call__(self):
                ...     return {"model": np.array([1,2]), "success": True}
                ...
                >>> inv_options = InversionOptions()
                >>> inv_options.set_tool(MyDummySolver)
                >>> inv_options.summary()
                =============================
                Summary for inversion options
                =============================
                Solving method: None set
                Use `suggest_solving_methods()` to check available solving methods.
                -----------------------------
                Backend tool: `<class '__main__.MyDummySolver'>` - My dummy solver that always return (1,2) as result
                References: []
                Use `suggest_tools()` to check available backend tools.
                -----------------------------
                Solver-specific parameters: None set
                Use `suggest_solver_params()` to check required/optional solver-specific parameters.

        """
        if tool is None:
            self.unset_tool()
        elif isinstance(tool, Type):
            if (
                issubclass(tool, Callable)
                and "__call__" not in tool.__abstractmethods__
            ):
                self.tool = tool
            else:
                raise ValueError(
                    "the custom solver class you've provided should implement"
                    " __call__(self,) methodread CoFI documentation 'tutorials -"
                    " Advanced Usage' section for how to plug inyour own solver"
                )
        else:
            if tool not in tool_dispatch_table:
                close_matches = difflib.get_close_matches(
                    tool, tool_dispatch_table.keys()
                )
                _error_msg_suffix = (
                    f"\n\nDid you mean '{close_matches[0]}'?"
                    if len(close_matches)
                    else ""
                )
                raise ValueError(
                    "the tool is invalid, please use"
                    " `InversionOptions.suggest_tools()` to see"
                    f" options{_error_msg_suffix}"
                )
            if self.method and tool not in tool_suggest_table[self.method]:
                warnings.warn(
                    f"the tool {tool} is valid but doesn't match the solving method"
                    f" you've selected: {self.method}"
                )
            self.tool = tool

    def unset_tool(self):
        """Unsets the chosen backend tool"""
        self.tool = None

    def get_tool(self) -> Union[str, Type]:
        """Get the backend tool chosen so far, or the default tool if not chosen

        Returns
        -------
        Union[str, Type]
            the name of the backend tool (if it's supported by us), or the class name
            of your own solver
        """
        return self.tool if self.tool else self.get_default_tool()

    def get_default_tool(self) -> str:
        """Get the default tool based on the chosen solving method

        Returns
        -------
        str
            the name of the default backend tool
        """
        if self.method:
            return tool_suggest_table[self.method][0]
        return tool_suggest_table["optimization"][0]

    def suggest_solving_methods(self):
        """Prints a list of solving methods to choose from

        Examples
        --------

        .. admonition:: code example
            :class: dropdown, attention

            .. code-block:: pycon
                :emphasize-lines: 3

                >>> from cofi import InversionOptions
                >>> inv_options = InversionOptions()
                >>> inv_options.suggest_solving_methods()
                The following solving methods are supported:
                {'optimization', 'matrix solvers'}

                Use `suggest_tools()` to see a full list of backend tools for each method

        """
        print("The following solving methods are supported:")
        print(solving_methods)
        print(
            "\nUse `suggest_tools()` to see a full list of backend tools for each"
            " method"
        )

    def suggest_tools(self):
        """Prints a list of tools based on the solving method chosen

        Examples
        --------

        .. admonition:: code example
            :class: dropdown, attention

            .. code-block:: pycon
                :emphasize-lines: 3, 15

                >>> from cofi import InversionOptions
                >>> inv_options = InversionOptions()
                >>> inv_options.suggest_tools()
                Here's a complete list of inversion tools supported by CoFI (grouped by methods):
                {
                    "optimization": [
                        "scipy.optimize.minimize",
                        "scipy.optimize.least_squares"
                    ],
                    "matrix solvers": [
                        "scipy.linalg.lstsq"
                    ]
                }
                >>> inv_options.set_solving_method("matrix solvers")
                >>> inv_options.suggest_tools()
                Based on the solving method you've set, the following tools are suggested:
                ['scipy.linalg.lstsq']

                Use `InversionOptions.set_tool(tool_name)` to set a specific tool from above
                Use `InversionOptions.set_solving_method(method_name)` to change solving method
                Use `InversionOptions.unset_solving_method()` if you'd like to see more options
                Check CoFI documentation 'Advanced Usage' section for how to plug in your own solver

        """
        if self.method:
            tools = tool_suggest_table[self.method]
            print(
                "Based on the solving method you've set, the following tools are"
                " suggested:"
            )
            print(tools)
            print(
                "\nUse `InversionOptions.set_tool(tool_name)` to set a specific tool"
                " from above"
            )
            print(
                "Use `InversionOptions.set_solving_method(method_name)` to change"
                " solving method"
            )
            print(
                "Use `InversionOptions.unset_solving_method()` if you'd like to see"
                " more options"
            )
            print(
                "Check CoFI documentation 'Advanced Usage' section for how to plug in"
                " your own tool or solver"
            )
        else:
            print(
                "Here's a complete list of inversion tools supported by CoFI (grouped"
                " by methods):"
            )
            print(json.dumps(tool_suggest_table, indent=4))

    def suggest_solver_params(self):
        """Prints required and optional solver-specific parameters

        Examples
        --------

        .. admonition:: code example
            :class: dropdown, attention

            .. code-block:: pycon
                :emphasize-lines: 3

                >>> from cofi import InversionOptions
                >>> inv_options = InversionOptions()
                >>> inv_options.suggest_solver_params()
                Current backend tool scipy.optimize.minimize (default) has the following solver-specific parameters:
                Required parameters:
                -- nothing --
                Optional parameters & default settings:
                {'method': None, 'tol': None, 'callback': None, 'options': None}

        """
        tool, dft_suffix = (
            (self.tool, "")
            if self.tool
            else (f"{self.get_default_tool()}", " (default)")
        )
        solver = tool_dispatch_table[tool]
        print(
            f"Current backend tool {tool}{dft_suffix} has the following solver-specific"
            " parameters:"
        )
        print("Required parameters:")
        print(
            solver.required_in_options()
            if solver.required_in_options()
            else "-- nothing --"
        )
        print("Optional parameters & default settings:")
        print(
            solver.optional_in_options()
            if solver.optional_in_options()
            else "-- nothing --"
        )

    def summary(self):
        """Helper method that prints a summary of current ``InversionOptions`` object
        to console

        Examples
        --------

        .. admonition:: code example
            :class: dropdown, attention

            .. code-block:: pycon
                :emphasize-lines: 3

                >>> from cofi import InversionOptions
                >>> inv_options = InversionOptions()
                >>> inv_options.summary()
                =============================
                Summary for inversion options
                =============================
                Solving method: None set
                Use `suggest_solving_methods()` to check available solving methods.
                -----------------------------
                Backend tool: `scipy.optimize.minimize (by default)` - SciPy's optimizers that minimizes a scalar function with respect to one or more variables, check SciPy's documentation page for a list of methods
                References: ['https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.minimize.html']
                Use `suggest_tools()` to check available backend tools.
                -----------------------------
                Solver-specific parameters: None set
                Use `suggest_solver_params()` to check required/optional solver-specific parameters.

        """
        self._summary()

    def _summary(self, display_lines=True):
        # inspiration from keras: https://keras.io/examples/vision/mnist_convnet/
        title = "Summary for inversion options"
        display_width = len(title)
        double_line = "=" * display_width
        single_line = "-" * display_width
        if display_lines:
            print(double_line)
        print(title)
        if display_lines:
            print(double_line)
        solving_method = self.method if self.method else "None set"
        tool, dft_suffix = (
            (self.tool, "")
            if self.tool
            else (f"{self.get_default_tool()}", " (by default)")
        )
        tool = tool_dispatch_table[tool] if isinstance(tool, str) else tool
        print(f"Solving method: {solving_method}")
        print("Use `suggest_solving_methods()` to check available solving methods.")
        if display_lines:
            print(single_line)
        print(f"Backend tool: `{tool}{dft_suffix}` - {tool.short_description}")
        # print(f"Backend tool description: {tool.short_description}")
        print(f"References: {tool.documentation_links}")
        print("Use `suggest_tools()` to check available backend tools.")
        if display_lines:
            print(single_line)
        params_suffix = "None set" if len(self.hyper_params) == 0 else ""
        print(f"Solver-specific parameters: {params_suffix}")
        for key, val in self.hyper_params.items():
            print(f"{key} = {val}")
        print(
            "Use `suggest_solver_params()` to check required/optional solver-specific"
            " parameters."
        )

    def __repr__(self) -> str:
        class_name = self.__class__.__name__
        method = f"'{self.method}'" if self.method else "unknown"
        tool = (
            f"'{self.tool}'" if self.tool else f"(default)'{self.get_default_tool()}'"
        )
        return f"{class_name}(method={method},tool={tool})"



import pytest

from cofi import BaseProblem
from cofi._exceptions import NotDefinedError


@pytest.fixture
def empty_problem():
    return BaseProblem()


def test_default_not_defined_error(empty_problem):
    with pytest.raises(NotDefinedError):
        empty_problem.objective(1)
    with pytest.raises(NotDefinedError):
        empty_problem.gradient(1)
    with pytest.raises(NotDefinedError):
        empty_problem.hessian(1)
    with pytest.raises(NotDefinedError):
        empty_problem.hessian_times_vector(1, 2)
    with pytest.raises(NotDefinedError):
        empty_problem.residual(1)
    with pytest.raises(NotDefinedError):
        empty_problem.jacobian(1)
    with pytest.raises(NotDefinedError):
        empty_problem.jacobian_times_vector(1, 2)
    with pytest.raises(NotDefinedError):
        empty_problem.data_misfit(1)
    with pytest.raises(NotDefinedError):
        empty_problem.regularization(1)
    with pytest.raises(NotDefinedError):
        empty_problem.regularization_matrix(1)
    with pytest.raises(NotDefinedError):
        empty_problem.forward(1)
    with pytest.raises(NotDefinedError):
        empty_problem.data
    with pytest.raises(NotDefinedError):
        empty_problem.data_covariance
    with pytest.raises(NotDefinedError):
        empty_problem.data_covariance_inv
    with pytest.raises(NotDefinedError):
        empty_problem.initial_model
    with pytest.raises(NotDefinedError):
        empty_problem.model_shape
    with pytest.raises(NotDefinedError):
        empty_problem.bounds
    with pytest.raises(NotDefinedError):
        empty_problem.constraints
    with pytest.raises(NotDefinedError):
        empty_problem.log_posterior(1)
    with pytest.raises(NotDefinedError):
        empty_problem.log_prior(1)
    with pytest.raises(NotDefinedError):
        empty_problem.log_likelihood(1)
    with pytest.raises(NotDefinedError):
        empty_problem.log_posterior_with_blobs(1)
    with pytest.raises(NotDefinedError):
        empty_problem.blobs_dtype


def test_default_not_defined_properties(empty_problem):
    assert not empty_problem.objective_defined
    assert not empty_problem.gradient_defined
    assert not empty_problem.hessian_defined
    assert not empty_problem.hessian_times_vector_defined
    assert not empty_problem.residual_defined
    assert not empty_problem.jacobian_defined
    assert not empty_problem.jacobian_times_vector_defined
    assert not empty_problem.data_misfit_defined
    assert not empty_problem.regularization_defined
    assert not empty_problem.regularization_matrix_defined
    assert not empty_problem.forward_defined
    assert not empty_problem.data_defined
    assert not empty_problem.data_covariance_defined
    assert not empty_problem.data_covariance_inv_defined
    assert not empty_problem.initial_model_defined
    assert not empty_problem.model_shape_defined
    assert not empty_problem.bounds_defined
    assert not empty_problem.constraints_defined
    assert not empty_problem.log_posterior_defined
    assert not empty_problem.log_prior_defined
    assert not empty_problem.log_likelihood_defined
    assert not empty_problem.log_posterior_with_blobs_defined
    assert not empty_problem.blobs_dtype_defined


def test_default_defined_components(empty_problem):
    assert len(empty_problem.defined_components()) == 0


def test_default_summary(empty_problem):
    empty_problem.summary()



import numpy as np

from ._base_inference_tool import error_handler
from ._scipy_lstsq import ScipyLstSq


class ScipySparseLstSq(ScipyLstSq):
    r"""Wrapper for Scipy's sparse linear system solvers as laid out in:
    https://docs.scipy.org/doc/scipy/reference/sparse.linalg.html#solving-linear-problems.

    These solvers include direct and iterative solvers:
    - :func:`scipy.sparse.linalg.spsolve`
    - :func:`scipy.sparse.linalg.spsolve_triangular`
    - :func:`scipy.sparse.linalg.bicg`
    - :func:`scipy.sparse.linalg.bicgstab`
    - :func:`scipy.sparse.linalg.cg`
    - :func:`scipy.sparse.linalg.cgs`
    - :func:`scipy.sparse.linalg.gmres`
    - :func:`scipy.sparse.linalg.lgmres`
    - :func:`scipy.sparse.linalg.minres`
    - :func:`scipy.sparse.linalg.qmr`
    - :func:`scipy.sparse.linalg.gcrotmk`
    - :func:`scipy.sparse.linalg.tfqmr`
    - :func:`scipy.sparse.linalg.lsqr`
    - :func:`scipy.sparse.linalg.lsmr`

    The usage of this tool is similar to ``scipy.linalg.lstsq`` and users can include
    data noise and regularization in the inversion process. More specifically, there
    are four cases:

    1. basic case, where neither data noise nor regularization is taken into account,
       we solve :math:`Gm=d`
    2. with uncertainty, we solve :math:`m=(G^TC_d^{-1}G)^{-1}G^TC_d^{-1}d`
    3. with Tikhonov regularization, :math:`m=(G^TG+\lambda L^TL)^{-1}G^Td`
    4. with both uncertainty and regularization,
       :math:`m=(G^TC_d^{-1}G+\lambda L^TL)^{-1}G^TC_d^{-1}d`

    where:

    - :math:`m` refers to inferred model (solution)
    - :math:`G` refers to the Jacobian of the forward operator
    - :math:`d` refers to the data vector
    - :math:`\lambda` refers to the regularization factor that adjusts the ratio of
      data misfit and regularization term
    - :math:`L` refers to the regularization matrix, and is usually chosen to be
      damping (zeroth order Tikhonov), roughening (first order Tikhonov), or smoothing
      (second order Tikhonov).
    """

    documentation_links = [
        "https://docs.scipy.org/doc/scipy/reference/sparse.linalg.html#solving-linear-problems"
    ]
    short_description = (
        "SciPy's sparse linear system solvers for solving linear problems"
    )

    @classmethod
    def required_in_problem(cls) -> set:
        return {"jacobian", "data"}

    @classmethod
    def optional_in_problem(cls) -> dict:
        return {
            "data_covariance_inv": None,
            "data_covariance": None,
            "regularization_matrix": None,
        }

    @classmethod
    def required_in_options(cls) -> set:
        return set()

    @classmethod
    def optional_in_options(cls) -> dict:
        return {
            "algorithm": "spsolve",
            "atol": 0.0,
            "algorithm_params": dict(),
            "with_uncertainty_if_possible": True,
            "with_tikhonov_if_possible": True,
        }

    @classmethod
    def available_algorithms(cls) -> set:
        return {
            "spsolve",
            "spsolve_triangular",
            "bicg",
            "bicgstab",
            "cg",
            "cgs",
            "gmres",
            "lgmres",
            "minres",
            "qmr",
            "gcrotmk",
            "tfqmr",
            "lsqr",
            "lsmr",
        }

    def __init__(self, inv_problem, inv_options):
        super().__init__(inv_problem, inv_options)
        self._components_used = list(self.required_in_problem())
        self._assign_args()  # defined in super class (_scipy_lstsq.ScipyLstSq)
        self._assign_args_without_reg_or_noise()
        self._assign_atol()
        self._validate_algorithm()

    def __call__(self) -> dict:
        raw_result = self._call_backend_tool()
        if self._params["algorithm"] in ["spsolve", "spsolve_triangular"]:
            res = {"success": True, "model": raw_result}
        elif self._params["algorithm"] == "lsqr":
            res = {
                "success": True,
                "model": raw_result[0],
                "istop": raw_result[1],
                "itn": raw_result[2],
                "r1norm": raw_result[3],
                "r2norm": raw_result[4],
                "anorm": raw_result[5],
                "acond": raw_result[6],
                "arnorm": raw_result[7],
                "xnorm": raw_result[8],
                "var": raw_result[9],
            }
        elif self._params["algorithm"] == "lsmr":
            res = {
                "success": True,
                "model": raw_result[0],
                "istop": raw_result[1],
                "itn": raw_result[2],
                "normr": raw_result[3],
                "normar": raw_result[4],
                "norma": raw_result[5],
                "conda": raw_result[6],
                "normx": raw_result[7],
            }
        else:
            res = {
                "success": raw_result[1] == 0,
                "model": raw_result[0],
                "info": raw_result[1],
            }
        return res

    def _assign_args_without_reg_or_noise(self):
        if (
            (not self._params["with_uncertainty"])
            and (not self._params["with_tikhonov"])
            and self._params["algorithm"] in ["spsolve", "spsolve_triangular"]
        ):
            self._a = self._G
            self._b = self._d

    def _assign_atol(self):
        if self._params["algorithm"] not in ["spsolve", "spsolve_triangular", "minres"]:
            self._params["algorithm_params"]["atol"] = self._params["atol"]

    def _validate_algorithm(self):
        if self._params["algorithm"] not in self.available_algorithms():
            raise ValueError(
                f"the algorithm you've chosen ({self._params['algorithm']}) "
                "is invalid. Please choose from the following: "
                f"{self.available_algorithms()}"
            )
        if not isinstance(self._params["algorithm_params"], dict):
            raise ValueError(
                "algorithm_params should be a dictionary, e.g. {'rtol': 1e-6}"
            )
        import scipy

        self.func = getattr(scipy.sparse.linalg, self._params["algorithm"])

    @error_handler(
        when="when solving the linear system equation",
        context="in the process of solving",
    )
    def _call_backend_tool(self):
        return self.func(A=self._a, b=self._b, **self._params["algorithm_params"])


# CoFI -> Parameter estimation -> Matrix based solvers -> Linear system solvers -> scipy.sparse.linalg -> spsolve
# description: Direct method that solves the sparse linear system Ax=b, where b may be a vector or a matrix
# documentation: https://docs.scipy.org/doc/scipy/reference/generated/scipy.sparse.linalg.spsolve.html

# CoFI -> Parameter estimation -> Matrix based solvers -> Linear system solvers -> scipy.sparse.linalg -> spsolve_triangular
# description: Direct method that solves the sparse linear system Ax=b, assuming A is a triangular matrix.
# documentation: https://docs.scipy.org/doc/scipy/reference/generated/scipy.sparse.linalg.spsolve_triangular.html

# CoFI -> Parameter estimation -> Matrix based solvers -> Linear system solvers -> scipy.sparse.linalg -> bicg
# description: Iterative method that solves the sparse linear system Ax=b using the BIConjugate Gradient method
# documentation: https://docs.scipy.org/doc/scipy/reference/generated/scipy.sparse.linalg.bicg.html

# CoFI -> Parameter estimation -> Matrix based solvers -> Linear system solvers -> scipy.sparse.linalg -> bicgstab
# description: Iterative method that solves the sparse linear system Ax=b using the BIConjugate Gradient STABilized method
# documentation: https://docs.scipy.org/doc/scipy/reference/generated/scipy.sparse.linalg.bicgstab.html

# CoFI -> Parameter estimation -> Matrix based solvers -> Linear system solvers -> scipy.sparse.linalg -> cg
# description: Iterative method that solves the sparse linear system Ax=b using the Conjugate Gradient method
# documentation: https://docs.scipy.org/doc/scipy/reference/generated/scipy.sparse.linalg.cg.html

# CoFI -> Parameter estimation -> Matrix based solvers -> Linear system solvers -> scipy.sparse.linalg -> cgs
# description: Iterative method that solves the sparse linear system Ax=b using the Conjugate Gradient Squared method
# documentation: https://docs.scipy.org/doc/scipy/reference/generated/scipy.sparse.linalg.cgs.html

# CoFI -> Parameter estimation -> Matrix based solvers -> Linear system solvers -> scipy.sparse.linalg -> gmres
# description: Iterative method that solves the sparse linear system Ax=b using the Generalized Minimal RESidual method
# documentation: https://docs.scipy.org/doc/scipy/reference/generated/scipy.sparse.linalg.gmres.html

# CoFI -> Parameter estimation -> Matrix based solvers -> Linear system solvers -> scipy.sparse.linalg -> lgmres
# description: Iterative method that solves the sparse linear system Ax=b using the LGMRES algorithm
# documentation: https://docs.scipy.org/doc/scipy/reference/generated/scipy.sparse.linalg.lgmres.html

# CoFI -> Parameter estimation -> Matrix based solvers -> Linear system solvers -> scipy.sparse.linalg -> minres
# description: Iterative method that solves the sparse linear system Ax=b using the MINimum RESidual method
# documentation: https://docs.scipy.org/doc/scipy/reference/generated/scipy.sparse.linalg.minres.html

# CoFI -> Parameter estimation -> Matrix based solvers -> Linear system solvers -> scipy.sparse.linalg -> qmr
# description: Iterative method that solves the sparse linear system Ax=b using the Quasi-Minimal Residual method
# documentation: https://docs.scipy.org/doc/scipy/reference/generated/scipy.sparse.linalg.qmr.html

# CoFI -> Parameter estimation -> Matrix based solvers -> Linear system solvers -> scipy.sparse.linalg -> qcrotmk
# description: Iterative method that solves the sparse linear system Ax=b using the flexible GCROT(m,k) algorithm
# documentation: https://docs.scipy.org/doc/scipy/reference/generated/scipy.sparse.linalg.qcrotmk.html

# CoFI -> Parameter estimation -> Matrix based solvers -> Linear system solvers -> scipy.sparse.linalg -> tfqmr
# description: Iterative method that solves the sparse linear system Ax=b using the Transpose-Free Quasi-Minimal Residual method
# documentation: https://docs.scipy.org/doc/scipy/reference/generated/scipy.sparse.linalg.tfqmr.html

# CoFI -> Parameter estimation -> Matrix based solvers -> Linear system solvers -> scipy.sparse.linalg -> lsqr
# description: Iterative method that finds the least-squares solution to a large, sparse, linear system of equations
# documentation: https://docs.scipy.org/doc/scipy/reference/generated/scipy.sparse.linalg.lsqr.html

# CoFI -> Parameter estimation -> Matrix based solvers -> Linear system solvers -> scipy.sparse.linalg -> lsmr
# description: Iterative method that solves sparse least-squares problems
# documentation: https://docs.scipy.org/doc/scipy/reference/generated/scipy.sparse.linalg.lsmr.html



from abc import abstractmethod, ABCMeta
import warnings

from .._exceptions import CofiError


class BaseInferenceTool(metaclass=ABCMeta):
    r"""Base class for backend inference tool wrappers

    This is the point where we connect ``cofi`` to other inversion libraries or
    code. We expose this as a part of ``cofi``'s public interface, to facilitate minimal
    effort to link ``cofi`` to your own inversion code or external libraries that aren't
    connected by us yet.

    To create your own inference tool, simply subclass :class:`BaseInferenceTool` and
    define the following methods & fields.

    .. admonition:: Example definition of a custom solver
        :class: dropdown, hint

        .. code-block:: pycon

            >>> from cofi.tools import BaseInferenceTool
            >>> class MyDummySolver(BaseInferenceTool):
            ...   short_description = "My dummy solver that always return (1,2) as result"
            ...   documentation_links = ["https://cofi.readthedocs.io/en/latest/api/generated/cofi.tools.BaseInferenceTool.html"]
            ...   @classmethod
            ...   def required_in_problem(cls): return ["objective", "gradient"]
            ...   @classmethod
            ...   def optional_in_problem(cls): return {"initial_model": [0,0]}
            ...   @classmethod
            ...   def required_in_options(cls): return []
            ...   @classmethod
            ...   def optional_in_options(cls): return {"method": "dummy"}
            ...   def __init__(self, inv_problem, inv_options):
            ...     super().__init__(inv_problem, inv_options)
            ...   def __call__(self):
            ...     return {"model": np.array([1,2]), "success": True}
            ...
            >>> from cofi import InversionOptions
            >>> inv_options = InversionOptions()
            >>> inv_options.set_tool(MyDummySolver)
            >>> inv_options.summary()
            =============================
            Summary for inversion options
            =============================
            Solving method: None set
            Use `suggest_solving_methods()` to check available solving methods.
            -----------------------------
            Backend tool: `<class '__main__.MyDummySolver'>` - My dummy solver that always return (1,2) as result
            References: ['https://cofi.readthedocs.io/en/latest/api/generated/cofi.tools.BaseInferenceTool.html']
            Use `suggest_tools()` to check available backend tools.
            -----------------------------
            Solver-specific parameters: None set
            Use `suggest_solver_params()` to check required/optional solver-specific parameters.

    .. rubric:: Minimal implementation

    Define the following minimally. Input validation will be performed automatically
    when a new instance is created, based on what you've returned for the first four
    class methods below.

    .. autosummary::
        BaseInferenceTool.required_in_problem
        BaseInferenceTool.optional_in_problem
        BaseInferenceTool.required_in_options
        BaseInferenceTool.optional_in_options
        BaseInferenceTool.__init__
        BaseInferenceTool.__call__

    .. rubric:: Displaying

    In addition, the following class variables help us display properly.
    Failure to include them won't cause an error, but may result in some information
    mismatch in displaying methods like :func:`cofi.Inversion.summary`.

    .. autosummary::
        BaseInferenceTool.short_description
        BaseInferenceTool.documentation_links

    .. rubric:: Make it more complete

    All backend inference tools in ``cofi`` also update the following field, and this
    will be displayed via :func:`cofi.Inversion.summary`. It's not required but good
    to keep track of this:

    .. autosummary::
        BaseInferenceTool.components_used

    :ref:`back to top <top_BaseInferenceTool>`

    """

    #: list: references about the backend solver. It helps ensure the audience
    #: understands what the backend solver does. The list can include the official
    #: documentation if the solver wraps an external library, or links to papers
    #: and other online resources explaining the approach
    #:
    #: For example, we use the following as the ``documentation_links`` for
    #: solver wrapping :func:`scipy.linalg.lstsq`, so that users can see the details
    #: of what's in the backend::
    #:
    #:     documentation_links = [
    #:         "https://docs.scipy.org/doc/scipy/reference/generated/scipy.linalg.lstsq.html",
    #:         "https://www.netlib.org/lapack/lug/node27.html",
    #:     ]
    documentation_links = list()
    #: str: a short introduction about the solver. This is for display purpose only
    short_description = str()

    def __init__(self, inv_problem, inv_options):
        """initialisation routine for the solver instance

        You will need to implement this in the subclass, and it's recommended to have
        the following line included::

            super().__init__(inv_problem, inv_options)


        What it does are:

        - Attaching the :class:`BaseProblem` and :class:`InversionOptions` objects to
          ``self``;
        - Validating both input objects based on the information in
          :meth:`required_in_problem`, :meth:`optional_in_problem`,
          :meth:`required_in_options`, and :meth:`optional_in_options`;
        - Assigning inversion tool parameters to ``self._params`` based on what are
          returned by class methods :meth:`required_in_options`,
          :meth:`optional_in_options`, and what are set by users in the
          :class:`InversionOptions` object;
        - Initializing the ``self._components_used`` dictionary based on what are
          returned by class methods :meth:`required_in_problem`,
          :meth:`optional_in_problem`, and what are defined by users in the
          :class:`BaseProblem` object. You can overwrite this if your solver is more
          specific in deciding what components get used.

        Alternatively (if you want), you can also define your own validation routines,
        then you don't have to call the ``__init__`` method defined in this super
        class, and don't have to add things to the fields.

        Parameters
        ----------
        inv_problem : BaseProblem
            an inversion problem setup
        inv_options : InversionOptions
            an object that defines how to run the inversion

        """
        self._inv_problem = inv_problem
        self._inv_options = inv_options
        self._validate_inv_problem()
        self._validate_inv_options()
        self._assign_options()  # assigns options to self._params
        self._update_components_used()  # update components to self._components_used

    @classmethod
    @abstractmethod
    def required_in_problem(cls) -> set:
        r"""a set of components required in :class:`BaseProblem` instance

        This is a standard part for a subclass of :class:`BaseInferenceTool` and helps
        validate input :class:`BaseProblem` instance
        """
        return set()

    @classmethod
    @abstractmethod
    def optional_in_problem(cls) -> dict:
        r"""a dictionary of components that are optional in :class:`BaseProblem`
        instance

        The keys stand for name of the components in ``BaseProblem``, and the values
        input :class:`BaseProblem` instance
        """
        return dict()

    @classmethod
    @abstractmethod
    def required_in_options(cls) -> set:
        """a set of solver-specific options required in :class:`InversionOptions`
        instance

        This is a standard part for a subclass of :class:`BaseInferenceTool` and helps
        validate input :class:`InversionOptions` instance
        """
        return set()

    @classmethod
    @abstractmethod
    def optional_in_options(cls) -> dict:
        """dict: a dictioanry of solver-specific options that are optional in
        :class:`InversionOptions` instance

        This is a standard part for a subclass of :class:`BaseInferenceTool` and helps
        validate input :class:`InversionOptions` instance
        """
        return dict()

    @abstractmethod
    def __call__(self) -> dict:
        """the method that calls the backend inversion routines

        This is an abstract method, meaning that you have to implement this on your
        own in the subclass, otherwise the definition of the subclass will cause an
        error directly.

        Returns
        -------
        dict
            a Python dictionary that has at least ``model``/``models`` and ``success`` as
            keys

        """
        raise NotImplementedError

    @property
    def components_used(self) -> set:
        r"""a set of strings describing what components defined in :class:`BaseProblem`
        are used in this solving process. This is typically the intersection of three
        sets: :func:`cofi.BaseProblem.all_components`,
        :func:`BaseInferenceTool.required_in_problem` and keys of
        :func:`BaseInferenceTool.optional_in_problem`
        """
        return self._components_used

    @property
    def inv_problem(self):
        r"""the inversion problem to be solved

        This is the first argument in the constructor of :class:`BaseInferenceTool`
        """
        return self._inv_problem

    @property
    def inv_options(self):
        r"""the inveersion settings

        This is the second argument in the constructor of :class:`BaseInferenceTool`
        """
        return self._inv_options

    def _validate_inv_problem(self):
        # check whether enough information from inv_problem is provided
        defined = self.inv_problem.defined_components()
        required = self.required_in_problem()
        if all({component in defined for component in required}):
            return True
        raise ValueError(
            f"you've chosen {self.__class__.__name__} to be your solving tool, but "
            "not enough information is provided in the BaseProblem object - "
            f"required: {required}; provided: {defined}"
        )

    def _validate_inv_options(self):
        # check whether inv_options matches current solver (correctness of dispatch) from callee
        # check whether required options are provided (algorithm-specific)
        defined = self.inv_options.get_params()
        required = self.required_in_options()
        optional = self.optional_in_options()
        all_required_are_defined = all({option in defined for option in required})
        if all_required_are_defined:
            defined_list = list(defined)
            defined_but_not_required_or_optional = [
                option not in optional and option not in required
                for option in defined_list
            ]
            if any(defined_but_not_required_or_optional):
                from itertools import compress

                items = list(
                    compress(defined_list, defined_but_not_required_or_optional)
                )
                warnings.warn(
                    "the following options are defined but not in parameter list for "
                    f"the chosen tool: {items}"
                )
            return True
        raise ValueError(
            f"you've chosen {self.__class__.__name__} to be your solving tool, but "
            "not enough information is provided in the InversionOptions object - "
            f"required: {required}; provided: {defined}"
        )

    def _assign_options(self):
        params = self.inv_options.get_params()
        self._params = dict()
        for opt in self.required_in_options():
            self._params[opt] = params[opt]
        for opt, val in self.optional_in_options().items():
            self._params[opt] = params.get(opt, val)

    def _update_components_used(self):
        self._components_used = list(self.required_in_problem())
        defined_in_problem = self.inv_problem.defined_components()
        optional_components_defined = set(defined_in_problem).union(
            self.optional_in_problem()
        )
        self._components_used.extend(list(optional_components_defined))

    def __repr__(self) -> str:
        return self.__class__.__name__


def error_handler(when, context):
    """Error handler for running inference tools"""

    def wrap_error_handler(func):
        def wrapped_func(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except Exception as e:
                raise CofiError(
                    f"error ocurred {when} ({context}). Check exception details "
                    "from message above.",
                ) from e

        return wrapped_func

    return wrap_error_handler



import numpy
import pytest

from cofi import BaseProblem
from cofi._exceptions import (
    InvalidOptionError, 
    DimensionMismatchError, 
    InvocationError, 
)


############### TEST auto generated methods ###########################################
def test_obj_from_dm_reg():
    inv_problem = BaseProblem()
    # test valid
    inv_problem.set_data_misfit(lambda x: x**2)
    inv_problem.set_regularization(lambda x: x)
    assert inv_problem.objective(1) == 2
    # test invalid
    inv_problem.set_data_misfit(lambda x: x[2])
    with pytest.raises(InvocationError, match=r".*exception while calling auto-generated objective.*"):
        inv_problem.objective(1)


def test_obj_from_dm():
    inv_problem = BaseProblem()
    # test invalid
    inv_problem.set_data_misfit(lambda x: x[2])
    with pytest.raises(InvocationError):
        inv_problem.objective(1)


def test_lp_from_ll_lp():
    inv_problem = BaseProblem()
    # test valid
    inv_problem.set_log_likelihood(lambda x: x**2)
    inv_problem.set_log_prior(lambda x: x)
    assert inv_problem.log_posterior(1) == 2
    # test invalid
    inv_problem.set_log_likelihood(lambda x: x[2])
    with pytest.raises(InvocationError):
        inv_problem.log_posterior(1)


def test_hessp_from_hess():
    inv_problem = BaseProblem()
    # test invalid
    inv_problem.set_hessian(numpy.array([1,2]))
    with pytest.raises(InvocationError):
        inv_problem.hessian_times_vector(0, numpy.array([1,2,3]))


def test_jacp_from_jac():
    inv_problem = BaseProblem()
    # test invalid
    inv_problem.set_jacobian(numpy.array([1,2]))
    with pytest.raises(InvocationError):
        inv_problem.jacobian_times_vector(0, numpy.array([1,2,3]))


def test_res_from_fwd_dt():
    inv_problem = BaseProblem()
    inv_problem.set_forward(lambda x: x**2)
    inv_problem.set_data(numpy.array([1,4,9]))
    with pytest.raises(InvocationError):
        inv_problem.residual(numpy.array([1,2]))


############### TEST function wrapper #################################################
def test_wrapping_objective():
    inv_problem = BaseProblem()
    # 1
    inv_problem.set_objective(lambda a,b,c: a+b+c, args=[1], kwargs={"c":2})
    assert inv_problem.objective(1) == 4
    # 2
    inv_problem.set_forward(lambda a,b,c=3,d=2: a+b+c+d, args=[3,2], kwargs={"d":1})
    assert inv_problem.forward(1) == 7

def test_not_overwriting_by_autogen():
    inv_problem = BaseProblem()
    # 1
    inv_problem.set_log_likelihood(lambda _: 1)
    inv_problem.set_log_prior(lambda _: 2)
    assert inv_problem.log_posterior_defined
    assert inv_problem.log_posterior(1) == 3
    # 2
    inv_problem.set_log_posterior(lambda _: 4)
    assert inv_problem.log_posterior(1) == 4
    inv_problem.set_log_prior(lambda _: 5)
    assert inv_problem.log_posterior(1) == 4

def test_set_reg_with_args():
    inv_problem = BaseProblem()
    from scipy import sparse
    A = sparse.csr_matrix([[1, 2, 0], [0, 0, 3], [4, 0, 5]])
    inv_problem.set_regularization(lambda m, A: A @ m.T @ m, args=[A])
    inv_problem.regularization(numpy.array([1,2,3]))

def test_invalid_func():
    inv_problem = BaseProblem()
    with pytest.raises(InvalidOptionError, match=".*invalid, please choose from the following.*"):
        inv_problem.set_forward(1)


############### TEST set from BaseProblem.__init__ ####################################
def test_init_set_fwd():
    _fwd = lambda x: x**2
    _fwd_w_args = lambda x, a: x**a
    p1 = BaseProblem(forward=_fwd)
    assert p1.forward_defined
    assert p1.forward(2) == 4
    p2 = BaseProblem(forwardd=_fwd)
    assert not p2.forward_defined
    p3 = BaseProblem(forward={"forward":_fwd_w_args, "args":[2]}, model_shape=(1,2))
    assert p3.forward_defined
    assert p3.forward(2) == 4
    assert p3.model_shape_defined
    assert p3.model_shape[1] == 2


############### TEST check_defined ####################################################
def test_check_defined():
    inv_problem = BaseProblem()
    # function defined by set_*
    inv_problem.set_objective(lambda a: a + 1)
    assert inv_problem.objective_defined
    # function defined by attaching
    inv_problem.forward = lambda x: x**2
    assert inv_problem.forward_defined
    # repr
    assert "BaseProblem" in str(inv_problem)
    inv_problem.name = "AnotherProblem"
    assert "AnotherProblem" in str(inv_problem)
    # initial model and model shape
    inv_problem.set_initial_model(numpy.array([1, 2, 3]))
    assert inv_problem.initial_model_defined
    assert inv_problem.model_shape_defined
    assert inv_problem.model_shape == (3,)
    with pytest.raises(DimensionMismatchError, match=r".*the model shape you've provided.*"):
        inv_problem.set_model_shape((2, 1))
    inv_problem.set_model_shape((3, 1))


############### TEST suggest_tools ##################################################
def test_suggest_tools(capsys):
    inv_problem = BaseProblem()
    # 0
    inv_problem.suggest_tools()
    console_output = capsys.readouterr().out
    assert "scipy.optimize.minimize" not in console_output
    assert "scipy.linalg.lstsq" not in console_output
    # 1
    inv_problem.set_initial_model(1)
    inv_problem.set_objective(lambda x: x)
    inv_problem.suggest_tools()
    console_output = capsys.readouterr().out
    assert "scipy.optimize.minimize" in console_output
    assert "scipy.optimize.least_squares" not in console_output
    assert "scipy.linalg.lstsq" not in console_output
    # 2
    inv_problem.set_jacobian(numpy.array([1]))
    inv_problem.set_data(2)
    inv_problem.suggest_tools()
    console_output = capsys.readouterr().out
    assert "scipy.linalg.lstsq" in console_output


def test_suggest_tools_return():
    inv_problem = BaseProblem()
    inv_problem.set_jacobian(numpy.array([1]))
    inv_problem.set_data(2)
    suggested = inv_problem.suggest_tools(print_to_console=False)
    assert "scipy.linalg.lstsq" in suggested["matrix solvers"]



import functools
import numpy as np

from . import BaseInferenceTool, error_handler


class ScipyLstSq(BaseInferenceTool):
    r"""Wrapper for generalised linear system solver :func:`scipy.linalg.lstsq`

    There are four cases:

    1. basic case, where neither data noise nor regularization is taken into account,
       :math:`m=(G^TG)^{-1}G^Td`
    2. with uncertainty, :math:`m=(G^TC_d^{-1}G)^{-1}G^TC_d^{-1}d`
    3. with Tikhonov regularization, :math:`m=(G^TG+\lambda L^TL)^{-1}G^Td`
    4. with both uncertainty and regularization,
       :math:`m=(G^TC_d^{-1}G+\lambda L^TL)^{-1}G^TC_d^{-1}d`

    where:

    - :math:`m` refers to inferred model (solution)
    - :math:`G` refers to the Jacobian of the forward operator
    - :math:`d` refers to the data vector
    - :math:`\lambda` refers to the regularization factor that adjusts the ratio of
      data misfit and regularization term
    - :math:`L` refers to the regularization matrix, and is usually chosen to be
      damping (zeroth order Tikhonov), roughening (first order Tikhonov), or smoothing
      (second order Tikhonov).
    """

    documentation_links = [
        "https://docs.scipy.org/doc/scipy/reference/generated/scipy.linalg.lstsq.html",
        "https://www.netlib.org/lapack/lug/node27.html",
    ]
    short_description = (
        "SciPy's wrapper function over LAPACK's linear least-squares solver, "
        "using 'gelsd', 'gelsy' (default), or 'gelss' as backend driver"
    )

    @classmethod
    def required_in_problem(cls) -> set:
        return _init_class_methods()[0]

    @classmethod
    def optional_in_problem(cls) -> dict:
        return _init_class_methods()[1]

    @classmethod
    def required_in_options(cls) -> set:
        return _init_class_methods()[2]

    @classmethod
    def optional_in_options(cls) -> dict:
        return _init_class_methods()[3]

    @classmethod
    def available_algorithms(cls) -> set:
        return {"gelsd", "gelsy", "gelss"}

    def __init__(self, inv_problem, inv_options):
        super().__init__(inv_problem, inv_options)
        self._components_used = list(self.required_in_problem())
        self._assign_args()

    def _assign_args(self):
        inv_problem = self.inv_problem

        # get jacobian matrix (presumably jacobian is a constant)
        if inv_problem.initial_model_defined:
            dummy_model = inv_problem.initial_model
        elif inv_problem.model_shape_defined:
            dummy_model = np.ones(inv_problem.model_shape)
        else:
            dummy_model = np.array([])
        try:
            self._G = inv_problem.jacobian(dummy_model)
        except Exception as exception:
            raise ValueError(
                "jacobian function isn't set properly for least squares solver, "
                "this should return a matrix unrelated to model vector"
            ) from exception

        # get data vector
        self._d = inv_problem.data

        # check whether to take uncertainty into account
        self._params["with_uncertainty"] = self._params[
            "with_uncertainty_if_possible"
        ] and (
            inv_problem.data_covariance_defined
            or inv_problem.data_covariance_inv_defined
        )
        # get Cd_inv if needed
        if self._params["with_uncertainty"]:
            if not inv_problem.data_covariance_inv_defined:
                self._Cd_inv = np.linalg.inv(inv_problem.data_covariance)
                self._components_used.append("data_covariance")
            else:
                self._Cd_inv = inv_problem.data_covariance_inv
                self._components_used.append("data_covariance_inv")
            # check diagonal (for potential shortcut in computation)
            diag_elem = np.diag(self._Cd_inv).copy()
            np.fill_diagonal(self._Cd_inv, 0)
            is_diagonal = (self._Cd_inv == 0).all()
            np.fill_diagonal(self._Cd_inv, diag_elem)
            if is_diagonal:
                _gt_cdinv = self._G.T * diag_elem
            else:
                _gt_cdinv = self._G.T @ self._Cd_inv
            self._a = _gt_cdinv @ self._G
            self._b = _gt_cdinv @ self._d
        else:
            self._a = self._G.T @ self._G
            self._b = self._G.T @ self._d

        # check whether to take regularization into account
        self._params["with_tikhonov"] = (
            self._params["with_tikhonov_if_possible"]
            and inv_problem.regularization_defined
        )
        # get lamda and L matrix if needed
        if self._params["with_tikhonov"]:
            if inv_problem.regularization_matrix_defined:
                try:
                    _L = inv_problem.regularization_matrix(dummy_model)
                except Exception as exception:
                    raise ValueError(
                        "regularization matrix function isn't set properly for least "
                        "squares solver, this should return a matrix unrelated to "
                        "model vector"
                    ) from exception
                self._LtL = _L.T @ _L
                self._components_used.append("regularization_matrix")
            else:
                self._LtL = np.identity(self._G.shape[1])
            self._a += self._LtL

    def __call__(self) -> dict:
        res_p, residual, rank, singular_vals = self._call_lstsq()
        res = {
            "success": True,
            "model": res_p,
            "sum_of_squared_residuals": residual,
            "effective_rank": rank,
            "singular_values": singular_vals,
        }
        if self._params["with_uncertainty"]:
            res["model_covariance"] = self._calculate_model_covariance()
        return res

    @error_handler(
        when="when solving the linear system equation",
        context="in the process of solving",
    )
    def _call_lstsq(self):
        from scipy.linalg import lstsq

        return lstsq(
            a=self._a,
            b=self._b,
            cond=self._params["cond"],
            overwrite_a=self._params["overwrite_a"],
            overwrite_b=self._params["overwrite_b"],
            check_finite=self._params["check_finite"],
            lapack_driver=self._params["lapack_driver"],
        )

    @error_handler(
        when="when calculating the model covariance estimation",
        context="after the solving process",
    )
    def _calculate_model_covariance(self):
        return np.linalg.inv(self._a)


@functools.lru_cache(maxsize=None)
def _init_class_methods():
    import inspect
    from scipy.linalg import lstsq

    _scipy_lstsq_args = dict(inspect.signature(lstsq).parameters)
    required_in_problem: set = {"jacobian", "data"}
    optional_in_problem: dict = {
        "data_covariance_inv": None,
        "data_covariance": None,
        "regularization_matrix": None,
    }
    required_in_options: set = set()
    optional_in_options: dict = {
        k: v.default for k, v in _scipy_lstsq_args.items() if k not in {"a", "b"}
    }
    optional_in_options["with_uncertainty_if_possible"] = True
    optional_in_options["with_tikhonov_if_possible"] = True
    return (
        required_in_problem,
        optional_in_problem,
        required_in_options,
        optional_in_options,
    )


# CoFI -> Parameter estimation -> Matrix based solvers -> Linear system solvers -> scipy.linalg.lstsq -> gelsd
# description: Solve linear least squares using divide-and-conquer SVD.

# CoFI -> Parameter estimation -> Matrix based solvers -> Linear system solvers -> scipy.linalg.lstsq -> gelsy
# description: Solve linear least squares using complete orthogonal factorization.

# CoFI -> Parameter estimation -> Matrix based solvers -> Linear system solvers -> scipy.linalg.lstsq -> gelss
# description: Solve linear least squares using SVD.



from typing import Type

from . import BaseProblem, InversionOptions
from .tools import tool_dispatch_table, BaseInferenceTool


class InversionResult:
    r"""The result class of an inversion run.

    You won't need to create an object of this class by yourself. See :func:`Inversion.run`
    for how you will get such an instance.

    .. seealso::

        When using sampling methods, you get a :class:`SamplingResult` object, with
        additional analysis methods attached.

    """

    #: bool: indicates status of the inversion run
    success: bool
    #: dict: raw output from backend ineference tools
    res: dict

    def __init__(self, res: dict) -> None:
        self.__dict__.update(res)
        self.res = res
        if "success" not in res:
            raise ValueError(
                "inversion termination status not returned in result dictionary, "
                "fix your solver to return properly. Check CoFI documentation "
                "'tutorial - Advanced Usage' section for how to plug in your own solver"
            )
        self.success_or_not = (
            "success" if hasattr(self, "success") and self.success else "failure"
        )

    def summary(self) -> None:
        """Helper method that prints a summary of the inversion result to console"""
        self._summary()

    def _summary(self, display_lines=True) -> None:
        title = "Summary for inversion result"
        display_width = len(title)
        double_line = "=" * display_width
        single_line = "-" * display_width
        if display_lines:
            print(double_line)
        print(title)
        if display_lines:
            print(double_line)
        print(self.success_or_not.upper())
        if display_lines:
            print(single_line)
        for key, val in self.res.items():
            if key != "success":
                if key == "sampler":
                    print(f"{key}: <{val.__module__}.{val.__class__.__name__} object>")
                else:
                    print(f"{key}: {val}")

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.success_or_not})"


class SamplingResult(InversionResult):
    """the result class of an inversion run, when the inversion is sampling-based

    This is a subclass of :class:`InversionResult`, so has the full functionality of
    it. Additionally, you can convert a :class:`SamplingResult` object into an
    :class:`arviz.InferenceData` object so that various plotting functionalities are
    available from arviz.

    """

    def __init__(self, res: dict) -> None:
        super().__init__(res)
        if "sampler" not in res:
            raise ValueError(
                "sampler not found in class SamplingResult, very likely to be a bug on"
                " our (CoFI) side. Please file an issue at "
                "https://github.com/inlab-geo/cofi/issues, thanks!"
            )

    def to_arviz(self, **kwargs):
        """convert sampler result into an :class:`arviz.InferenceData` object

        Note that this method takes in keyword arguments that matches the
        ``arviz.from_<library>`` function. If your results are sampled from emcee,
        then you can pass in any keyword arguments as described in
        :func:`arviz.from_emcee`.

        Returns
        -------
        arviz.InferenceData
            an :class:`arviz.InferenceData` object converted from your sampler

        Raises
        ------
        NotImplementedError
            when sampling result of current type (``type(SamplingResult.sampler)``))
            cannot be converted into an :class:`arviz.InferenceData`
        """
        import arviz
        import emcee

        sampler = self.sampler
        if sampler is None:
            raise ValueError(
                "sampling result is None, please double check that your solver returns"
                " correctly if you are using your own solver; otherwise please file an"
                " issue at https://github.com/inlab-geo/cofi/issues, thanks!"
            )
        if isinstance(sampler, emcee.EnsembleSampler):
            if (
                hasattr(self, "blob_names")
                and self.blob_names
                and "blob_names" not in kwargs
            ):
                if "blob_groups" in kwargs:
                    self.arviz_inference_data = arviz.from_emcee(
                        sampler,
                        blob_names=self.blob_names,
                        **kwargs,
                    )
                else:
                    blobs_groups = [
                        ("prior" if name == "log_prior" else name)
                        for name in self.blob_names
                    ]  # "log_prior" isn't in arviz's supported groups, use "prior"
                    self.arviz_inference_data = arviz.from_emcee(
                        sampler,
                        blob_names=self.blob_names,
                        blob_groups=blobs_groups,
                        **kwargs,
                    )
            else:
                self.arviz_inference_data = arviz.from_emcee(sampler, **kwargs)
            return self.arviz_inference_data
        else:
            raise NotImplementedError(
                f"sampling result of type {sampler.__class__.__name__} not supported"
                " yet, please file an issue at https://github.com/inlab-geo/cofi/issues"
                ", thanks!"
            )


class Inversion:
    r"""The class holder that take in both an inversion problem setup :class:`BaseProblem`
    and inversion solver options :class:`InversionOptions`, and handles the running of an
    inversion.

    Recall that we have 4 main steps to define and run an inversion through ``cofi``:

    1. Define a :class:`BaseProblem` object
    2. Define an :class:`InversionOptions` object
    3. Pass both of the above objects into an :class:`Inversion`
    4. Hit that :func:`Inversion.run` method and get the result :class:`InversionResult`

    So let's think of ``Inversion`` as an engine that manages the input and output of an
    inversion run for you.

    .. admonition:: Example usage of Inversion
        :class: attention

        >>> from cofi import BaseProblem, InversionOptions, Inversion
        >>> inv_problem = BaseProblem()
        >>> inv_problem.set_... # attach info about your problem
        >>> inv_options = InversionOptions()
        >>> inv_options.set_... # select backend tool and solver-specific parameters
        >>> inv = Inversion(inv_problem, inv_options)
        >>> inv_result = inv.run()

        See our example gallery for more inversion runs.

    .. admonition:: A future direction?
        :class: seealso

        We seperate out this "inversion", instead of passing a ``BaseProblem`` object
        directly to a hypothetical ``InversionSolver`` concept. This is not only because
        we want a cleaner workflow, but also because we imagine this ``Inversion``
        object to have more capability::

        >>> inv = Inversion(inv_problem, inv_options)
        >>> inv_result = inv.run()
        >>> inv.save("filename")
        >>> inv = Inversion.load("filename")
        >>> inv.analyse("filename")

    """

    def __init__(self, inv_problem: BaseProblem, inv_options: InversionOptions) -> None:
        self.inv_problem = inv_problem
        self.inv_options = inv_options
        # dispatch inversion_solver from self.inv_options, validation is done by solver
        self.inv_solve = self._dispatch_solver()(inv_problem, inv_options)
        self.inv_result = None

    def run(self) -> InversionResult:
        """Starts the inversion and returns an :class:`InversionResult` object.

        The inversion will be entirely based on the setup defined in ``BaseProblem`` and
        ``InversionOptions`` objects.

        Returns
        -------
        InversionResult
            the result of inversion that has attributes ``model`` / ``models`` and ``success``
            minimally. Check :class:`InversionResult` for details.
        """
        res_dict = self.inv_solve()
        if "sampler" in res_dict:
            self.inv_result = SamplingResult(res_dict)
        else:
            self.inv_result = InversionResult(res_dict)
        return self.inv_result

    def _dispatch_solver(self) -> Type[BaseInferenceTool]:
        tool = self.inv_options.get_tool()
        # look up tool_dispatch_table to return constructor for a BaseInferenceTool subclass
        if isinstance(tool, str):
            return tool_dispatch_table[tool]
        # self-defined BaseInferenceTool (note that a BaseInferenceTool object is a callable)
        return self.inv_options.tool

    def summary(self):
        r"""Helper method that prints a summary of current ``Inversion`` object
        to console

        This is essentially a higher level method that calls the ``.summary()`` method
        on all of the three objects:

        - :class:`InversionResult` (if the inversion has finished)
        - :class:`BaseProblem`
        - :class:`InversionOptions`

        """
        title = "Summary for Inversion"
        subtitle_result = "Completed with the following result:"
        subtitle_options = "With inversion solver defined as below:"
        subtitle_problem = "For inversion problem defined as below:"
        display_width = max(
            len(title),
            len(subtitle_result),
            len(subtitle_options),
            len(subtitle_problem),
        )
        double_line = "=" * display_width
        single_line = "-" * display_width
        print(double_line)
        print(title)
        print(double_line)
        if self.inv_result:
            print(f"{subtitle_result}\n")
            self.inv_result._summary(False)
            print(single_line)
        else:
            print("Inversion hasn't started, try `inversion.run()` to see result")
            print(single_line)
        print(f"{subtitle_options}\n")
        self.inv_options._summary(False)
        print(single_line)
        print(f"{subtitle_problem}\n")
        self.inv_problem._summary(False)
        if self.inv_result:
            print("List of functions/properties got used by the backend tool:")
            print(self.inv_solve.components_used)



import functools
import numpy as np

from . import BaseInferenceTool, error_handler


class Emcee(BaseInferenceTool):
    documentation_links = [
        "https://emcee.readthedocs.io/en/stable/user/sampler/#emcee.EnsembleSampler",
        "https://emcee.readthedocs.io/en/stable/user/sampler/#emcee.EnsembleSampler.sample",
    ]
    short_description = (
        "an MIT licensed pure-Python implementation of Goodman & Weare’s Affine "
        "Invariant Markov chain Monte Carlo (MCMC) Ensemble sampler"
    )

    @classmethod
    def required_in_problem(cls) -> set:
        return _init_class_methods()[0]

    @classmethod
    def optional_in_problem(cls) -> dict:
        return _init_class_methods()[1]

    @classmethod
    def required_in_options(cls) -> set:
        return _init_class_methods()[2]

    @classmethod
    def optional_in_options(cls) -> dict:
        return _init_class_methods()[3]

    @classmethod
    @functools.lru_cache(maxsize=None)
    def available_algorithms(cls) -> set:
        from emcee import moves

        return {move for move in moves.__all__ if move != "Move"}

    def __init__(self, inv_problem, inv_options):
        super().__init__(inv_problem, inv_options)
        self._assign_args()
        self._initialize_sampler()

    def __call__(self) -> dict:
        self._run_mcmc()
        result = {
            "success": True,
            "sampler": self.sampler,
            "blob_names": self._params["blob_names"],
        }
        return result

    def _assign_args(self):
        # assign components in problem to args
        inv_problem = self.inv_problem
        self._params["blobs_dtype"] = None
        self._params["blob_names"] = None
        if inv_problem.log_posterior_with_blobs_defined:
            self._params["log_prob_fn"] = inv_problem.log_posterior_with_blobs
            if inv_problem.blobs_dtype_defined:
                self._params["blob_names"] = [
                    name for (name, _) in inv_problem.blobs_dtype
                ]
                # uncomment below once this has been fixed:
                # issue: https://github.com/arviz-devs/arviz/issues/2036
                # self._blobs_dtype = inv_problem._blobs_dtype
        else:
            self._params["log_prob_fn"] = inv_problem.log_posterior
        self._params["ndim"] = np.shape(self._params["initial_state"])[0]
        self._params["ndim"] = np.prod(inv_problem.model_shape)

    @error_handler(
        when="in creating emcee.EnsembleSampler object",
        context="before sampling",
    )
    def _initialize_sampler(self):
        from emcee import EnsembleSampler

        self.sampler = EnsembleSampler(
            nwalkers=self._params["nwalkers"],
            ndim=self._params["ndim"],
            log_prob_fn=self._params["log_prob_fn"],
            pool=self._params["pool"],
            moves=self._params["moves"],
            args=None,  # already handled by BaseProblem
            kwargs=None,  # already handled by BaseProblem
            backend=self._params["backend"],
            vectorize=self._params["vectorize"],
            blobs_dtype=self._params["blobs_dtype"],
            parameter_names=self._params["parameter_names"],
            a=self._params["a"],
            postargs=self._params["postargs"],
            threads=self._params["threads"],
            live_dangerously=self._params["live_dangerously"],
            runtime_sortingfn=self._params["runtime_sortingfn"],
        )

    @error_handler(
        when="when running sampling",
        context="in the process of sampling",
    )
    def _run_mcmc(self):
        self.sampler.reset()
        self.sampler.run_mcmc(
            initial_state=self._params["initial_state"],
            nsteps=self._params["nsteps"],
            log_prob0=self._params["log_prob0"],
            rstate0=self._params["rstate0"],
            blobs0=self._params["blobs0"],
            tune=self._params["tune"],
            skip_initial_state_check=self._params["skip_initial_state_check"],
            thin_by=self._params["thin_by"],
            thin=self._params["thin"],
            store=self._params["store"],
            progress=self._params["progress"],
            progress_kwargs=self._params["progress_kwargs"],
        )


@functools.lru_cache(maxsize=None)
def _init_class_methods():
    import inspect
    from emcee import EnsembleSampler

    _emcee_EnsembleSampler_args = dict(inspect.signature(EnsembleSampler).parameters)
    _emcee_EnsembleSampler_sample_args = dict(
        inspect.signature(EnsembleSampler.sample).parameters
    )
    required_in_problem = {"log_posterior"}
    optional_in_problem = dict()
    required_in_options = {"nwalkers", "nsteps", "initial_state"}
    optional_in_options = {
        k: v.default
        for k, v in _emcee_EnsembleSampler_args.items()
        if k not in {"nwalkers", "ndim", "log_prob_fn", "self", "args", "kwargs"}
    }
    optional_in_options.update({
        k: v.default
        for k, v in _emcee_EnsembleSampler_sample_args.items()
        if k not in {"iterations", "self"}
    })
    return (
        required_in_problem,
        optional_in_problem,
        required_in_options,
        optional_in_options,
    )


# CoFI -> Ensemble methods -> Bayesian sampling -> McMC samplers -> emcee -> MHMove
# description: A general Metropolis-Hastings proposal.
# documentation: https://emcee.readthedocs.io/en/stable/user/moves/#emcee.moves.MHMove

# CoFI -> Ensemble methods -> Bayesian sampling -> McMC samplers -> emcee -> GaussianMove
# description: A Metropolis step with a Gaussian proposal function.
# documentation: https://emcee.readthedocs.io/en/stable/user/moves/#emcee.moves.GaussianMove

# CoFI -> Ensemble methods -> Bayesian sampling -> McMC samplers -> emcee -> RedBlueMove
# description: An abstract red-blue ensemble move with parallelization as described in Foreman-Mackey et al. (2013).
# documentation: https://emcee.readthedocs.io/en/stable/user/moves/#emcee.moves.RedBlueMove

# CoFI -> Ensemble methods -> Bayesian sampling -> McMC samplers -> emcee -> StretchMove
# description: A Goodman & Weare (2010) “stretch move” with parallelization as described in Foreman-Mackey et al. (2013).
# documentation: https://emcee.readthedocs.io/en/stable/user/moves/#emcee.moves.StretchMove

# CoFI -> Ensemble methods -> Bayesian sampling -> McMC samplers -> emcee -> WalkMove
# description: A Goodman & Weare (2010) “walk move” with parallelization as described in Foreman-Mackey et al. (2013).
# documentation: https://emcee.readthedocs.io/en/stable/user/moves/#emcee.moves.WalkMove

# CoFI -> Ensemble methods -> Bayesian sampling -> McMC samplers -> emcee -> KDEMove
# description: A proposal using a KDE of the complementary ensemble.
# documentation: https://emcee.readthedocs.io/en/stable/user/moves/#emcee.moves.KDEMove

# CoFI -> Ensemble methods -> Bayesian sampling -> McMC samplers -> emcee -> DEMove
# description: A proposal using differential evolution.
# documentation: https://emcee.readthedocs.io/en/stable/user/moves/#emcee.moves.DEMove

# CoFI -> Ensemble methods -> Bayesian sampling -> McMC samplers -> emcee -> DESnookerMove
# description: A snooker proposal using differential evolution.
# documentation: https://emcee.readthedocs.io/en/stable/user/moves/#emcee.moves.DESnookerMove



from typing import Union
from numbers import Number
import numpy as np
from scipy import sparse

from ._reg_base import BaseRegularization
from .._exceptions import DimensionMismatchError


REG_TYPES = {
    "damping": 0,
    "flattening": 1,
    "roughening": 1,
    "smoothing": 2,
}


class LpNormRegularization(BaseRegularization):
    r"""CoFI's utility class to calculate Lp-norm regularization, given the p value
    (default to 2), an optional weighting matrix and an optional reference value

    :math:`L(p, W, m_0) = ||W(m-m_0)||_p^p = \sum_i |W(m-m_0)_i|^p`

    With element-wise gradient of:

    :math:`\sum_i p|(W(m-m_0))_i|^{p-1}sign((W(m-m_0))_i)W_{ij}`

    And element-wise hessian of:

    :math:`\sum_i p(p-1)|(W(m-m_0))_i|^{p-2}W_{ij}W_{ik}`

    Where :math:`W` is a weighting matrix either generated given a specified type
    (e.g. :code:`weighting_matrix="smoothing"`), or a bring-your-own matrix
    (e.g. :code:`weighting_matrix=my_matrix`). This weighting matrix is by default
    in sparse type :class:`scipy.sparse.csr_matrix`.

    The weighting matrix (if not bring-your-own) can be generated provided with an
    option from {:code:`"damping"`, :code:`"flattening"`, :code:`"smoothing"`}.

    - If ``weighting_matrix == "damping"``, then

      .. toggle::

        - :attr:`matrix` is the identity matrix of size :math:`(M,M)`, where
          :math:`M` is the number of model parameters

    - If ``weighting_matrix == "roughening"`` (or equivalently ``"flattening"``),
      then

      .. toggle::

        - :attr:`matrix` is :math:`W` that we generate based on the model shape you've
          provided. We use the Python package :mod:`findiff` to generate it.
          "By default, findiff uses finite difference schemes with second
          order accuracy in the grid spacing." (See `findiff documentation on
          Derivatives <https://findiff.readthedocs.io/en/latest/source/getstarted.html#accuracy-control>`_
          for more details). For 1D problems, it looks like

          :math:`\begin{pmatrix}-1.5&2&-0.5&&&\\-0.5&&0.5&&&&&\\&-0.5&&0.5&&&&\\&&...&&...&&&\\&&&-0.5&&0.5&\\&&&&-0.5&&0.5\\&&&&0.5&-2&1.5\end{pmatrix}`

          While for higher dimension problems, by default it's a flattened version of
          an N-D array. The actual ordering of model parameters in higher dimensions
          is controlled by :class:`findiff.operators.FinDiff`.

    - If ``reg_type == "smoothing"``, then

      .. toggle::

        - :attr:`matrix` is :math:`W` that we generate based on the model shape you've
          provided. We use the Python package :mod:`findiff` to generate it.
          "By default, findiff uses finite difference schemes with second
          order accuracy in the grid spacing." (See `findiff documentation on
          Derivatives <https://findiff.readthedocs.io/en/latest/source/getstarted.html#accuracy-control>`_
          for more details). For 1D problems, it looks like

          :math:`\begin{pmatrix}2&-5&4&-1&&&\\1&-2&1&&&&\\&1&-2&1&&&\\&&...&...&...&&\\&&&1&-2&1&\\&&&&1&-2&1\\&&&-1&4&-5&2\end{pmatrix}`

          .. :math:`\begin{pmatrix}1&-2&1&&&&\\&1&-2&1&&&\\&&...&...&...&&\\&&&1&-2&1&\\&&&&1&-2&1\end{pmatrix}`

          While for higher dimension problems, by default it's a flattened version of
          an N-D array. The actual ordering of model parameters in higher dimensions
          is controlled by :class:`findiff.operators.FinDiff`.

    Parameters
    ----------
    p : Number
        order value (p in the formula above), default to 2
    weighting_matrix: str or np.ndarray
        regularization type (one of {:code:`"damping"`, :code:`"flattening"`
        :code:`"smoothing"`}), or a bring-your-own weighting matrix, default to
        "damping" (identity matrix for weighting)
    model_shape: tuple
        shape of the model, must be supplied if the :code:`reference_model` is not
        given
    reference_model: np.ndarray
        :math:`m_0` in the formula above

    Raises
    ------
    ValueError
        if neither :code:`model_size` nor :code:`reference_model` is given
    DimensionMismatchError
        if both :code:`model_size` and :code:`reference_model` are given but they don't
        match in dimension

    Examples
    --------

    Generate an L1 norm damping regularization for models of size 3:

    >>> from cofi.utils import LpNormRegularization
    >>> my_reg = LpNormRegularization(p=1, model_shape=(3,))
    >>> my_reg(np.array([0,2,1]))
    3.0

    To use togethter with :class:`cofi.BaseProblem`:

    >>> from cofi import BaseProblem
    >>> my_problem.set_regularization(my_reg)
    """

    def __init__(
        self,
        p: Number = 2,
        weighting_matrix: Union[str, np.ndarray] = "damping",
        model_shape: tuple = None,
        reference_model: np.ndarray = None,
    ):
        self._order = self._validate_p(p)
        self._weighting_matrix = weighting_matrix
        self._model_shape = self._validate_shape(model_shape, reference_model)
        self._reference_model = reference_model
        self._generate_weighting_matrix()

    def reg(self, model: np.ndarray) -> Number:
        flat_m = self._validate_model(model)
        diff_m = self._model_diff_to_ref(flat_m)
        weighted_diff_m = self._weighting_matrix @ diff_m
        return self._lp_norm(weighted_diff_m)

    def gradient(self, model: np.ndarray) -> np.ndarray:
        flat_m = self._validate_model(model)
        diff_m = self._model_diff_to_ref(flat_m)
        weighted_diff_m = self._weighting_matrix @ diff_m
        grad_lp_norm = self._lp_norm_gradient(weighted_diff_m)
        return self.matrix.T @ grad_lp_norm

    def hessian(self, model: np.ndarray) -> np.ndarray:
        W = self._weighting_matrix
        flat_m = self._validate_model(model)
        diff_m = self._model_diff_to_ref(flat_m)
        weighted_diff_m = W @ diff_m
        hess_lp_norm = self._lp_norm_hessian(weighted_diff_m)
        return W.T @ np.diag(hess_lp_norm) @ W

    @property
    def model_shape(self) -> tuple:
        return self._model_shape

    @property
    def matrix(self) -> sparse.csr_matrix:
        """the regularization matrix

        This is either an identity matrix, or first/second order difference matrix
        (generated by Python package ``findiff``), or a custom matrix brought on your
        own.
        """
        return self._weighting_matrix

    def _generate_weighting_matrix(self):
        import findiff

        if (
            isinstance(self._weighting_matrix, str)
            and self._weighting_matrix in REG_TYPES
        ) or self._weighting_matrix is None:
            _reg_type = self._weighting_matrix
            if _reg_type == "damping" or _reg_type is None:  # 0th order difference
                self._weighting_matrix = sparse.identity(self.model_size, format="csr")
            elif _reg_type in REG_TYPES:  # 1st/2nd order difference
                if np.size(self.model_shape) == 1:  # 1D model
                    order = REG_TYPES[_reg_type]
                    if self.model_size < order + 2:
                        raise ValueError(
                            f"the model_size needs to be at least >={order+2} "
                            f"for regularization type '{_reg_type}'"
                        )
                    d_dx = findiff.FinDiff(0, 1, order)
                    self._weighting_matrix = d_dx.matrix((self.model_size,))
                elif (
                    np.size(self.model_shape) == 2 and np.ndim(self.model_shape) == 1
                ):  # 2D model
                    nx = self.model_shape[0]
                    ny = self.model_shape[1]
                    order = REG_TYPES[_reg_type]
                    if nx < order + 2 or ny < order + 2:
                        raise ValueError(
                            f"the model_size needs to be at least (>={order+2},"
                            f" >={order+2}) for regularization type '{_reg_type}'"
                        )
                    d_dx = findiff.FinDiff(0, 1, order)  # x direction
                    d_dy = findiff.FinDiff(1, 1, order)  # y direction
                    matx = d_dx.matrix((nx, ny))  # scipy sparse matrix
                    maty = d_dy.matrix((nx, ny))  # scipy sparse matrix
                    self._weighting_matrix = np.vstack(
                        (matx.toarray(), maty.toarray())
                    )  # combine above
                else:
                    raise NotImplementedError(
                        "only 1D and 2D derivative operators implemented"
                    )
        elif is_matrix_like(self._weighting_matrix):  # byo matrix
            if len(self._weighting_matrix.shape) != 2:
                raise ValueError(
                    "the bring-your-own regularization matrix must be 2-dimensional"
                )
            elif self._weighting_matrix.shape[1] != self.model_size:
                raise ValueError(
                    "the bring-your-own regularization matrix must be in shape (_, M) "
                    "where M is the model size"
                )
            self._weighting_matrix = sparse.csr_matrix(self._weighting_matrix)
        else:
            raise ValueError(
                "please specify the weighting matrix either via a string among "
                "\{`damping`, `flattening`, `smoothing`\}, or bringing your own matrix"
            )

    @staticmethod
    def _validate_p(p):
        if not isinstance(p, Number):
            raise ValueError(
                f"number expected for argument `p` but got {p} of type {type(p)}"
            )
        elif p <= 0:
            raise ValueError(f"positive number expected for argument `p` but got {p}")
        return p

    @staticmethod
    def _validate_shape(model_shape, reference_model):
        if model_shape is None and reference_model is None:
            raise ValueError("please provide the model shape")
        elif model_shape is None and reference_model is not None:
            return reference_model.shape
        elif model_shape is not None and reference_model is None:
            if not isinstance(model_shape, tuple):
                raise TypeError(
                    "expected model shape in tuple (e.g. `(100,)`) but got "
                    f"{model_shape} instead"
                )
            return model_shape
        else:
            if reference_model.shape != model_shape:
                try:
                    np.reshape(reference_model, model_shape)
                except:
                    raise DimensionMismatchError(
                        entered_dimension=reference_model.shape,
                        entered_name="reference_model",
                        expected_dimension=model_shape,
                        expected_source="model_shape",
                    )
            return model_shape

    def _validate_model(self, model):
        flat_m = np.ravel(model)
        if flat_m.size != self.model_size:
            raise DimensionMismatchError(
                entered_name="model",
                entered_dimension=model.shape,
                expected_source="model_size",
                expected_dimension=self.model_size,
            )
        return flat_m

    def _model_diff_to_ref(self, model):
        if self._reference_model is None:
            return model
        else:
            return model - np.ravel(self._reference_model)

    def _lp_norm(self, mat):
        return np.sum(np.abs(mat) ** self._order)

    def _lp_norm_gradient(self, mat):
        return self._order * np.abs(mat) ** (self._order - 1) * np.sign(mat)

    def _lp_norm_hessian(self, mat):
        p = self._order
        return p * (p - 1) * np.abs(mat) ** (p - 2)


class QuadraticReg(LpNormRegularization):
    r"""CoFI's utility class to calculate weighted L2 norm regularization with an
    optional reference model

    :math:`L(W, m_0) = ||D(m-m_0)||_2^2`

    With gradient of:

    :math:`2\times W(m-m_0)`

    And hessian of:

    :math:`2\times W.T W`

    Where :math:`W` is a weighting matrix either generated given a specified type
    (e.g. :code:`weighting_matrix="smoothing"`), or a bring-your-own matrix
    (e.g. :code:`weighting_matrix=my_matrix`). This weighting matrix is by default
    in sparse type :class:`scipy.sparse.csr_matrix`.

    This class is a special case of :class:`LpNormRegularization` with :code:`p=2`.

    Parameters
    ----------
    weighting_matrix: str or np.ndarray
        regularization type (one of {:code:`"damping"`, :code:`"flattening"`
        :code:`"smoothing"`}), or a bring-your-own weighting matrix, default to
        "damping" (identity matrix for weighting)
    model_shape : tuple
        shape of the model, must be supplied if the :code:`reference_model` is not
        given
    reference_model: np.ndarray
        :math:`m_0` in the formula above

    Raises
    ------
    ValueError
        if neither :code:`model_size` nor :code:`reference_model` is given
    DimensionMismatchError
        if both :code:`model_size` and :code:`reference_model` are given but they don't
        match in dimension

    Examples
    --------

    Generate an quadratic smoothing regularization for models of size 3:

    >>> from cofi.utils import QuadraticReg
    >>> my_reg = QuadraticReg(weighting_matrix="smoothing", model_shape=(4,))
    >>> my_reg(np.array([0,2,1,0]))
    53.99999999999999

    To use togethter with :class:`cofi.BaseProblem`:

    >>> from cofi import BaseProblem
    >>> my_problem = BaseProblem()
    >>> my_problem.set_regularization(my_reg)
    """

    def __init__(
        self,
        weighting_matrix: Union[str, np.ndarray] = "damping",
        model_shape: tuple = None,
        reference_model: np.ndarray = None,
    ):
        super().__init__(
            p=2,
            weighting_matrix=weighting_matrix,
            model_shape=model_shape,
            reference_model=reference_model,
        )


matrix_like_classes = [np.ndarray] + [
    getattr(sparse, name) for name in sparse.__all__ if name.endswith("_matrix")
]


def is_matrix_like(obj):
    return any(isinstance(obj, cls) for cls in matrix_like_classes)



from numpy import argmin


from . import BaseInferenceTool, error_handler
from ._neighpyI import NeighpyI
from ._neighpyII import NeighpyII


class Neighpy(BaseInferenceTool):
    r"""Wrapper for the tool :code:`neighpy`, a Python implementation of the Neighbourhood Algorithm.

    The Neighbourhood Algorithm is a direct search method that uses a set of points in the parameter space to explore the objective function. It is a derivative-free method that is particularly useful for high-dimensional problems.

    Split into two phases, the direct search phase and the appraisal phase, this wrapper runs both phases.
    """

    documentation_links = ["https://neighpy.readthedocs.io/en/latest/index.html"]
    short_description = [
        "Wrapper for Neighpy, a Python implementation of the Neighbourhood Algorithm"
    ]

    @classmethod
    def required_in_problem(cls) -> set:
        # nothing required in problem from NeighpyII
        return NeighpyI.required_in_problem()

    @classmethod
    def optional_in_problem(cls) -> dict:
        return {**NeighpyI.optional_in_problem(), **NeighpyII.optional_in_problem()}

    @classmethod
    def required_in_options(cls) -> set:
        # only need n_resample and n_walkers from NeighpyII
        # as we will use the output of the direct search phase
        # as the initial ensemble
        return NeighpyI.required_in_options() | {"n_resample", "n_walkers"}

    @classmethod
    def optional_in_options(cls) -> dict:
        return {**NeighpyI.optional_in_options(), **NeighpyII.optional_in_options()}

    def __init__(self, inv_problem, inv_options):
        super().__init__(inv_problem, inv_options)
        self._params["ndim"] = len(self._params["bounds"])
        self._components_used = list(self.required_in_problem())

    def __call__(self) -> dict:
        raw_results = self._call_backend_tool()
        _best = raw_results["direct_search_samples"][
            argmin(raw_results["direct_search_objectives"])
        ]
        res = {"success": True, "model": _best}
        return {**res, **raw_results}

    @error_handler(
        when="at some point in the Neighbourhood Algorithm",
        context="",
    )
    def _call_backend_tool(self):
        searcher = NeighpyI._initialise_searcher(self)
        direct_search_samples, direct_search_objectives = NeighpyI._call_searcher(
            searcher, parallel=not self._params["serial"]
        )

        # a bit hacky but this is how we pass the results to the appraiser
        self._params["initial_ensemble"] = direct_search_samples
        self._params["log_ppd"] = -direct_search_objectives

        appraiser = NeighpyII._initalise_appraiser(self)
        appraisal_samples = NeighpyII._call_appriaser(appraiser)

        result = {
            "direct_search_samples": direct_search_samples,
            "direct_search_objectives": direct_search_objectives,
            "appraisal_samples": appraisal_samples,
        }

        return result


# CoFI -> Ensemble methods -> Direct search -> Monte Carlo -> Neighpy -> Neighbourhood Algorithm
# description: Wrapper for the tool Neighpy, implementing the Neighbourhood Algorithm
# documentation: https://neighpy.readthedocs.io/en/latest/



import numpy as np
import scipy

from . import BaseInferenceTool


class CoFISimpleNewton(BaseInferenceTool):
    documentation_links = [
        "https://en.wikipedia.org/wiki/Newton%27s_method_in_optimization",
    ]
    short_description = (
        "CoFI's own solver - simple Newton's approach (for testing mainly)"
    )

    @classmethod
    def required_in_problem(cls) -> set:
        return {"objective", "gradient", "hessian", "initial_model"}

    @classmethod
    def optional_in_problem(cls) -> dict:
        return dict()

    @classmethod
    def required_in_options(cls) -> set:
        return {"num_iterations"}

    @classmethod
    def optional_in_options(cls) -> dict:
        return {
            "step_length": 1,
            "verbose": True,
            "hessian_is_symmetric": False,
            "obj_tol": 1e-6,  # tolerance for change in objective function
            "param_tol": 1e-6,  # tolerance for change in model parameters
            "max_iterations": None,  # alias for num_iterations
        }

    def __init__(self, inv_problem, inv_options):
        super().__init__(inv_problem, inv_options)
        self._params["num_iterations"] = (
            self._params["max_iterations"] or self._params["num_iterations"]
        )

    def __call__(self) -> dict:
        m = self.inv_problem.initial_model
        m_prev = np.copy(m)
        obj_val = self.inv_problem.objective(m)
        obj_val_prev = obj_val
        self._n_obj_evaluations = 1
        self._n_grad_evaluations = 0
        self._n_hess_evaluations = 0

        for i in range(self._params["num_iterations"]):
            m = m + self._calculate_step(m)
            obj_val = self._verbose_objective_value(i, m)
            stop = self._stopping_criteria(obj_val, obj_val_prev, m, m_prev)
            if stop:
                break
            m_prev = m.copy()
            obj_val_prev = obj_val

        return {
            "model": m,
            "num_iterations": i,
            "objective_val": obj_val,
            "success": True,
            "n_obj_evaluations": self._n_obj_evaluations,
            "n_grad_evaluations": self._n_grad_evaluations,
            "n_hess_evaluations": self._n_hess_evaluations,
        }

    def _calculate_step(self, m):
        grad = self.inv_problem.gradient(m)
        self._n_grad_evaluations += 1
        hess = np.atleast_2d((self.inv_problem.hessian(m)))
        self._n_hess_evaluations += 1
        if self._params["hessian_is_symmetric"]:
            hess = scipy.sparse.csr_matrix(hess)
            step = scipy.sparse.linalg.minres(hess, -grad)[0]
        else:
            step = scipy.linalg.solve(hess, -grad)
        step = np.squeeze(np.asarray(step))
        step *= self._params["step_length"]
        return step

    def _stopping_criteria(self, obj_val, obj_val_prev, m, m_prev):
        # checking change in objective function value
        if np.abs(obj_val - obj_val_prev) < self._params["obj_tol"]:
            if self._params["verbose"]:
                print("Change in objective function below tolerance, stopping.")
            return True
        # checking biggest change in model parameters
        if np.max(np.abs(m - m_prev)) < self._params["param_tol"]:
            if self._params["verbose"]:
                print("Change in model parameters below tolerance, stopping.")
            return True
        return False

    def _verbose_objective_value(self, i, m):
        obj_val = self.inv_problem.objective(m)
        self._n_obj_evaluations += 1
        if self._params["verbose"]:
            print(f"Iteration #{i}, updated objective function value: {obj_val}")
        return obj_val


# CoFI -> Parameter estimation -> Optimization -> Non linear -> cofi.simple_newton -> Newton's method in optimization
# description: CoFI's own implementation of the Newton's method in optimization with stopping criteria.
# documentation: https://en.wikipedia.org/wiki/Newton%27s_method_in_optimization



import pytest
import numpy as np
import emcee

from cofi import SamplingResult, InversionResult


def test_validation():
    # 1 - InversionResult
    with pytest.raises(ValueError, match=r".*status not returned.*"):
        res = InversionResult(dict())
    # 2 - SamplingResult
    with pytest.raises(ValueError, match=r".*sampler not found.*"):
        res = SamplingResult({"success": True})

def test_success_status():
    # 1
    res = InversionResult({"success": True})
    assert res.success_or_not == "success"
    # 2
    res = InversionResult({"success": 1})
    assert res.success_or_not == "success"
    # 3
    res = InversionResult({"success": False})
    assert res.success_or_not == "failure"
    # 4
    res = InversionResult({"success": 0})
    assert res.success_or_not == "failure"

def test_summary(capsys):
    res = SamplingResult({"success": True, "sampler": emcee.EnsembleSampler(32,4,lambda x:x)})
    res.summary()
    console_output = capsys.readouterr().out
    assert "0x" not in console_output
    assert "emcee.ensemble.EnsembleSampler" in console_output

def test_to_arviz():
    # 1 - None sampler
    res = SamplingResult({"success": True, "sampler": None})
    with pytest.raises(ValueError):
        inf_data = res.to_arviz()
    # 2 - incorrect sampler type
    res = SamplingResult({"success": True, "sampler": np.array([1])})
    with pytest.raises(NotImplementedError):
        inf_data = res.to_arviz()
    # 3 - correct sampler (posterior)
    def dummy_pdf(model):
        if model < 0 or model > 1:
            return -np.inf
        return 0.0 # model lies within bounds -> return log(1)
    log_prob = dummy_pdf
    sampler = emcee.EnsembleSampler(2,1,log_prob)
    sampler.run_mcmc(np.array([[0.1],[0.3]]), 100)
    res = SamplingResult({"success": True, "sampler": sampler})
    idata = res.to_arviz()
    # 4 - correct sampler (prior + likelihood)
    def log_prob(model):
        pdf = dummy_pdf(model)
        return pdf+pdf, pdf, pdf
    sampler = emcee.EnsembleSampler(2,1,log_prob)
    sampler.run_mcmc(np.array([[0.1],[0.3]]), 100)
    res = SamplingResult({"success": True, "sampler": sampler, "blob_names": ["ll","lp"]})
    ## it is not clear to me what this was meant to test I was unable to find the string
    ## fragment 'group is not defined' anywhere in the cofi source code and no warning
    ## is being raised by the test give this is identified as a correct sampler on line 56.
    ## So for the time being the pytest.warns is removed. JRH
    #with pytest.warns(UserWarning, match=".*group is not defined.*"):
    idata = res.to_arviz()
    # 5 - correct sampler (prior + likelihood + blob_groups)
    idata = res.to_arviz(blob_groups=["log_likelihood", "prior"])



import functools

from . import BaseInferenceTool, error_handler


class BayesBay(BaseInferenceTool):
    r"""Wrapper for the tool BayesBay, a trans-dimensional reversible jump Bayesian
    sampling framework
    """

    documentation_links = [
        "https://bayes-bay.readthedocs.io/en/latest/api/generated/bayesbay.BaseBayesianInversion.html#bayesbay.BaseBayesianInversion"
    ]
    short_description = [
        "Wrapper for BayesBay, a trans-dimensional reversible jump Bayesian sampling"
        "framework"
    ]

    @classmethod
    def required_in_problem(cls) -> set:
        return set()

    @classmethod
    def optional_in_problem(cls) -> dict:
        return {"log_likelihood": None}

    @classmethod
    def required_in_options(cls) -> set:
        return {"perturbation_funcs", "walkers_starting_states"}

    @classmethod
    def optional_in_options(cls) -> dict:
        return _inspect_default_options()

    @classmethod
    def available_algorithms(cls) -> set:
        import bayesbay as bb

        return {s for s in bb.samplers.__all__ if s != "Sampler"}

    def __init__(self, inv_problem, inv_options):
        super().__init__(inv_problem, inv_options)
        self._components_used = list(self.required_in_problem())
        self._initialize_bb_inversion()

    def __call__(self) -> dict:
        self._call_backend_tool()
        samples = self._bb_bayes_inversion.get_results()
        res = {
            "success": True,
            "models": samples,
            "sampler": self._bb_bayes_inversion,
        }
        return res

    @error_handler(
        when="when initializating bayesbay.BaseBayesianInversion",
        context="in the process of preparing",
    )
    def _initialize_bb_inversion(self):
        _log_like_defined = self.inv_problem.log_likelihood_defined
        assert not (
            not _log_like_defined and self._params["log_like_ratio_func"] is None
        ), "at least one of `log_likelihood` and `log_like_ratio_func` to be defined"

        import bayesbay as bb

        self._bb_bayes_inversion = bb.BaseBayesianInversion(
            walkers_starting_states=self._params["walkers_starting_states"],
            perturbation_funcs=self._params["perturbation_funcs"],
            perturbation_weights=self._params["perturbation_weights"],
            log_like_ratio_func=self._params["log_like_ratio_func"],
            log_like_func=(
                self.inv_problem.log_likelihood if _log_like_defined else None
            ),
            n_chains=self._params["n_chains"],
            save_dpred=self._params["save_dpred"],
        )

    @error_handler(
        when="when calling bayesbay.BaseBayesianInversion.run method",
        context="in the process of solving",
    )
    def _call_backend_tool(self):
        self._bb_bayes_inversion.run(
            sampler=self._params["sampler"],
            n_iterations=self._params["n_iterations"],
            burnin_iterations=self._params["burnin_iterations"],
            save_every=self._params["save_every"],
            verbose=self._params["verbose"],
            print_every=self._params["print_every"],
            parallel_config=self._params["parallel_config"],
        )


@functools.lru_cache(maxsize=None)
def _inspect_default_options():
    import inspect
    from bayesbay import BaseBayesianInversion

    _bb_inv_init_args = dict(inspect.signature(BaseBayesianInversion).parameters)
    optional_in_options: dict = {
        k: v.default
        for k, v in _bb_inv_init_args.items()
        if v.default is not inspect._empty and k != "log_likelihood_func"
    }
    _bb_inv_run_args = dict(inspect.signature(BaseBayesianInversion.run).parameters)
    optional_in_options.update({
        k: v.default
        for k, v in _bb_inv_run_args.items()
        if v.default is not inspect._empty
    })
    return optional_in_options


# CoFI -> Ensemble methods -> Bayesian sampling -> Trans-D McMC -> bayesbay -> VanillaSampler
# description: Sampling the posterior by means of reversible-jump Markov chain Monte Carlo.
# documentation: https://bayes-bay.readthedocs.io/en/latest/api/generated/bayesbay.samplers.VanillaSampler.html

# CoFI -> Ensemble methods -> Bayesian sampling -> Trans-D McMC -> bayesbay -> ParallelTempering
# description: Sampling the posterior by means of reversible-jump Markov chain Monte Carlo accelerated with parallel tempering.
# documentation: https://bayes-bay.readthedocs.io/en/latest/api/generated/bayesbay.samplers.ParallelTempering.html

# CoFI -> Ensemble methods -> Bayesian sampling -> Trans-D McMC -> bayesbay -> SimulatedAnnealing
# description: Sampling the posterior by means of reversible-jump Markov chain Monte Carlo accelerated with simulated annealing.
# documentation: https://bayes-bay.readthedocs.io/en/latest/api/generated/bayesbay.samplers.SimulatedAnnealing.html



r"""Utility classes and functions (e.g. to generate regularization terms and more)

The class inheritance of regularization classes:

.. mermaid::

    graph TD;
    BaseRegularization --> LpNormRegularization;
    LpNormRegularization --> QuadraticReg;
    BaseRegularization --> ModelCovariance;
    ModelCovariance --> GaussianPrior;

"""

from ._reg_base import BaseRegularization
from ._reg_lp_norm import LpNormRegularization, QuadraticReg
from ._reg_model_cov import ModelCovariance, GaussianPrior

from ._multiple_runs import InversionPool


__all__ = [
    "BaseRegularization",
    "LpNormRegularization",
    "QuadraticReg",
    "ModelCovariance",
    "GaussianPrior",
    "InversionPool",
]



from typing import Callable, List, Union, Tuple
import multiprocessing
import itertools

from .._base_problem import BaseProblem
from .._inversion_options import InversionOptions
from .._inversion import Inversion, InversionResult


class InversionPool:
    """This class manages an ensemble of inversions and allows them to be run in parallel or sequentially.

    Parameters
    ----------
    list_of_inv_problems : Union[List[BaseProblem], BaseProblem]
        A list of problems for which the inversions need to be performed, or a single problem. Each problem should be an instance of the BaseProblem class.

    list_of_inv_options : Union[List[InversionOptions], InversionOptions]
        A list of options for the inversions, or a single set of options. Each set of options should be an instance of the InversionOptions class.

    callback : Callable[[InversionResult, int], None], optional
        A callback function that gets invoked for each problem after its inversion has been performed. It should accept two arguments - the result of the inversion (an instance of InversionResult) and the index of the problem. The callback can return a result that will be collected and returned in the callback results list. If no callback function is provided, None will be used.

    parallel : bool, optional
        A flag that determines whether the inversions should be performed in parallel or sequentially. If True, the inversions will be performed in parallel using multiple processes. If False or not provided, the inversions will be performed sequentially in the same process.

    Examples
    --------

    An example of using this function to make an L-curve:

    >>> import cofi
    >>> import matplotlib.pyplot as plt
    >>> def my_objective(m, lamda):
    ...     data_misfit = numpy.sum((m * data_x**2 - data_y) ** 2)
    ...     regularization = lamda * numpy.sum(m**2)
    ...     return data_misfit + regularization
    ...
    >>> my_problems = []
    >>> my_lamdas = numpy.logspace(-8, 8, 30)
    >>> for lamda in my_lamdas:
    ...     my_problem = cofi.BaseProblem()
    ...     my_problem.set_objective(my_objective, args=(lamda,))
    ...     my_problem.set_initial_model(100)
    ...     my_problems.append(my_problem)
    ...
    >>> my_options = cofi.InversionOptions()
    >>> my_options.set_tool("scipy.optimize.minimize")
    >>> def my_callback(inv_result, i):
    ...     m = inv_result.model
    ...     data_misfit = numpy.linalg.norm(m * data_x**2 - data_y)
    ...     model_norm = numpy.linalg.norm(m)
    ...     return data_misfit, model_norm
    ...
    >>> my_ensemble_of_inversions = cofi.utils.InversionPool(my_problems, my_options, my_callback, False)
    >>> all_results, all_cb_returns = my_ensemble_of_inversions.run()
    >>> l_curve_points = list(zip(*all_cb_returns))
    >>> plt.plot(l_curve_points)
    """

    def __init__(
        self,
        list_of_inv_problems: Union[List[BaseProblem], BaseProblem],
        list_of_inv_options: Union[List[InversionOptions], InversionOptions],
        callback: Callable[[InversionResult, int], None] = None,
        parallel: bool = False,
    ):
        if not list_of_inv_problems:
            raise ValueError(
                "empty list detected, please pass in a concrete list of BaseProblem "
                "instances"
            )
        if not list_of_inv_options:
            raise ValueError(
                "empty list detected, please pass in a concrete list of"
                " InversionOptions instances"
            )
        if (
            isinstance(list_of_inv_problems, list)
            and isinstance(list_of_inv_options, list)
            and len(list_of_inv_problems) != len(list_of_inv_options)
        ):
            raise ValueError(
                "mismatch in lengths, `list_of_inv_problems` has length "
                f"{len(list_of_inv_problems)} and `list_of_inv_options has length "
                f"{len(list_of_inv_options)}. Please make sure they are of the same "
                "length"
            )
        self.problems = (
            list_of_inv_problems
            if isinstance(list_of_inv_problems, list)
            else itertools.repeat(list_of_inv_problems)
        )
        self.options = (
            list_of_inv_options
            if isinstance(list_of_inv_options, list)
            else itertools.repeat(list_of_inv_options)
        )
        self.callback = callback
        self.parallel = parallel

    def run(self):
        """
        Runs all the inversions in the ensemble, either in parallel or sequentially.

        Returns
        -------
        Tuple[List[InversionResult], List]
            A tuple containing two lists - the first list contains the results of the inversions (as InversionResult objects) for each problem in the same order as the input problems list. The second list contains the results returned by the callback function for each problem in the same order as the input problems list.
        """
        if self.parallel:
            return self._run_multiple_inversions_parallel(
                self.problems, self.options, self.callback
            )
        else:
            return self._run_multiple_inversions_sequential(
                self.problems, self.options, self.callback
            )

    def _run_multiple_inversions_sequential(
        self,
        list_of_inv_problems: List[BaseProblem],
        list_of_inv_options: List[InversionOptions],
        callback: Callable[[InversionResult, int], None] = None,
    ) -> Tuple[List[InversionResult], List]:
        results = []
        for i, (problem, inv_options) in enumerate(
            zip(list_of_inv_problems, list_of_inv_options)
        ):
            res, callback_res = self._run_one_inversion_with_callback(
                problem, inv_options, i, callback
            )
            results.append((res, callback_res))
        results, callback_results = zip(*results)
        return results, callback_results

    def _run_multiple_inversions_parallel(
        self,
        list_of_inv_problems: List[BaseProblem],
        list_of_inv_options: List[InversionOptions],
        callback: Callable[[InversionResult, int], None] = None,
    ) -> Tuple[List[InversionResult], List]:
        callback_list = itertools.repeat(callback)
        i_s = range(len(list_of_inv_problems))
        all_args = zip(list_of_inv_problems, list_of_inv_options, i_s, callback_list)
        with multiprocessing.Pool() as pool:
            results = list(pool.imap(self._run_one_inversion_with_callback, all_args))
        results, callback_results = zip(*results)
        return results, callback_results

    @staticmethod
    def _run_one_inversion_with_callback(
        problem, inv_options=None, i=None, callback=None
    ):
        # for parallel case where only one argument is passed in,
        # unpack the first argument
        if not isinstance(problem, BaseProblem):
            problem, inv_options, i, callback = problem
        inv = Inversion(problem, inv_options)
        result = inv.run()
        callback_result = None
        if callback:
            callback_result = callback(result, i)
        return result, callback_result



import numpy as np
import pytest

from cofi import Inversion, BaseProblem, InversionOptions
from cofi.tools import BaseInferenceTool


@pytest.fixture
def polynomial_problem():
    inv_problem = BaseProblem()
    _x = np.array([1, 2, 3, 4, 5])
    _G = np.array([_x**i for i in range(3)]).T
    _m_true = np.array([2, 1, 1])
    _y = _G @ _m_true
    inv_problem.set_data(_y)
    inv_problem.set_jacobian(_G)
    inv_problem.set_jacobian_times_vector(lambda m, x: _G @ x)
    inv_problem.set_hessian(_G.T @ _G)
    inv_problem.set_hessian_times_vector(lambda m, x: _G.T @ _G @ x)
    inv_options = InversionOptions()
    inv_options.set_tool("scipy.linalg.lstsq")
    return inv_problem, inv_options


def test_solve(polynomial_problem):
    runner = Inversion(*polynomial_problem)
    inv_result = runner.run()
    assert inv_result.success


def test_runner_result_summary(polynomial_problem, capsys):
    runner = Inversion(*polynomial_problem)
    # 0
    runner.summary()
    console_output = capsys.readouterr()
    assert "scipy.linalg.lstsq" in console_output.out
    assert "Inversion hasn't started" in console_output.out
    # 1
    inv_result = runner.run()
    assert "success" in str(inv_result)
    inv_result.summary()
    console_output = capsys.readouterr()
    assert "SUCCESS" in console_output.out
    assert "model" in console_output.out
    runner.summary()
    console_output = capsys.readouterr()
    assert "SUCCESS" in console_output.out


def test_dispatch_custom_solver(polynomial_problem, capsys):
    inv_problem, inv_options = polynomial_problem

    class CustomSolver(BaseInferenceTool):
        def __call__(self) -> dict:
            return {"successful": True, "prob_optional": self.optional_in_problem()}

        @classmethod
        def required_in_problem(cls) -> set:
            return set()

        @classmethod
        def optional_in_problem(cls) -> dict:
            return dict()

        @classmethod
        def required_in_options(cls) -> set:
            return set()

        @classmethod
        def optional_in_options(cls) -> dict:
            return dict()

    inv_options.set_tool(CustomSolver)
    runner = Inversion(inv_problem, inv_options)
    with pytest.raises(ValueError):
        res = runner.run()
